import datamol as dm


def test_brics():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    frags = dm.fragment.brics(mol)
    assert len(frags) == 9


def test_frag():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    frags = dm.fragment.frag(mol)
    assert len(frags) == 9


def test_recap():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    frags = dm.fragment.recap(mol)
    assert len(frags) == 3


def test_anybreak():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    frags = dm.fragment.anybreak(mol)
    assert len(frags) == 9


def test_mmpa():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)

    frags = dm.fragment.mmpa_cut(mol)
    assert len(frags) == 39
    assert "CCCOCc1cccc(-c2ccccn2)c1,C(C[*:2])[*:1],C[*:1].c1ccc(-c2cccc(CO[*:2])c2)nc1\n" in frags


def test_assemble():
    # Fragment a molecule
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    frags = dm.fragment.brics(mol)

    # Limit the number of fragments to work with because
    # assembling is computationally intensive.
    frags = frags[:2]

    # Assemble molecules from the list of fragments
    mols = list(dm.fragment.assemble_fragment_order(frags, max_n_mols=4))

    assert len(mols) == 4


def test_break_mol():
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)
    fragments, *_, tree = dm.fragment.break_mol(mol, randomize=False, mode="brics", returnTree=True)

    assert fragments == ["CCC", "O", "C", "c1ccncc1", "c1ccccc1"]
    assert list(tree.nodes) == [0, 1, 2, 3, 4, 5, 6, 7, 8]
    assert list(tree.edges) == [(0, 1), (0, 2), (2, 3), (2, 4), (4, 5), (4, 6), (6, 7), (6, 8)]


def test_assemble_build():
    mols = [[dm.to_mol("CCCO"), dm.to_mol("CCCCCCCO")], [dm.to_mol("CCC"), dm.to_mol("CCCCCCC")]]

    results = list(dm.fragment.build(mols))
    assert len(results) == 71

    results = list(dm.fragment.build(mols, mode="rxn"))
    assert len(results) == 0

    results = list(dm.fragment.build(mols, mode=None))
    assert len(results) == 0



from ._structural import IsomerEnumerator

from ._enumerate import enumerate_stereoisomers
from ._enumerate import enumerate_tautomers
from ._enumerate import enumerate_structisomers
from ._enumerate import remove_stereochemistry
from ._enumerate import canonical_tautomer



from typing import Optional

from rdkit.Chem import Draw

import datamol as dm


def prepare_mol_for_drawing(mol: Optional[dm.Mol], kekulize: bool = True) -> Optional[dm.Mol]:
    """Prepare the molecule before drawing to avoid any error due to unsanitized molecule
    or incorrect valence or aromaticity.

    Code is inspired from `rdkit.Chem.Draw._moltoimg`.

    Args:
        mol: A molecule to prepare. If set to None, the function will return None.
        kekulize: Whether to kekulize the molecule.
    """

    if mol is None:
        return None

    try:
        with dm.without_rdkit_log():
            # Check for implicit and explicit valence
            if mol.NeedsUpdatePropertyCache():  # type: ignore
                mol.UpdatePropertyCache(False)  # type: ignore

            # Check for aromaticity
            if dm.is_lower_than_current_rdkit_version("2022.09"):
                _kekulize = Draw._okToKekulizeMol(mol, kekulize)
            else:
                _kekulize = Draw.shouldKekulize(mol, kekulize)

            # Run the rdkit preparation procedure
            _mol = Draw.rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=_kekulize)

    except ValueError:  # <- can happen on a kekulization failure
        # Run the rdkit preparation procedure with kekulize set to `False`
        _mol = Draw.rdMolDraw2D.PrepareMolForDrawing(mol, kekulize=False)

    return _mol



# type: ignore
import pytest

import numpy as np
import datamol as dm
import random


def test_generate():
    with pytest.raises(ValueError):
        smiles = "CCCC"
        mol = dm.to_mol(smiles)
        mol = dm.conformers.generate(mol, method="custom_method")

    smiles = "CCCC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None, minimize_energy=False)
    assert mol.GetNumConformers() == 50
    assert mol.GetConformer(0).GetPositions().shape == (4, 3)

    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(
        mol, rms_cutoff=None, minimize_energy=False, embed_params={"useRandomCoords": True}
    )
    assert mol.GetNumConformers() == 50
    assert mol.GetConformer(0).GetPositions().shape == (4, 3)


# We disable this test as it takes too long due to the size of the SMILES
@pytest.mark.skip
def test_generate_fail():
    # This mol should fail
    smiles = "C=CC1=C(N)Oc2cc1c(-c1cc(C(C)O)cc(=O)cc1C1NCC(=O)N1)c(OC)c2OC"
    mol = dm.to_mol(smiles)
    assert dm.conformers.generate(mol, n_confs=1, verbose=True, ignore_failure=True) is None


def test_generate_2():
    smiles = "CCCC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None, minimize_energy=True)
    assert mol.GetNumConformers() == 50
    assert "rdkit_UFF_energy" in mol.GetConformer(0).GetPropsAsDict()


def test_generate_3():
    smiles = "CCCC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=1, minimize_energy=False)
    assert mol.GetNumConformers() == 22
    assert mol.GetConformer(0).GetPositions().shape == (4, 3)


def test_generate_4():
    smiles = "CCCC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=1, minimize_energy=True)
    assert mol.GetNumConformers() == 21
    assert "rdkit_UFF_energy" in mol.GetConformer(0).GetPropsAsDict()


@pytest.mark.skip_platform("win")
def test_sasa():
    with pytest.raises(ValueError):
        smiles = "O=C(C)Oc1ccccc1C(=O)O"
        mol = dm.to_mol(smiles)
        mol = dm.conformers.sasa(mol)

    smiles = "CCCC=O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, minimize_energy=True)
    sasa = dm.conformers.sasa(mol)
    assert sasa.shape == (50,)


def test_rmsd():
    with pytest.raises(ValueError):
        smiles = "O=C(C)Oc1ccccc1C(=O)O"
        mol = dm.to_mol(smiles)
        mol = dm.conformers.rmsd(mol)

    smiles = "CCCC=O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None, minimize_energy=True)
    rmsd = dm.conformers.rmsd(mol)
    assert rmsd.shape == (50, 50)


def test_cluster():
    # no centroids
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None)
    clustered_mol = dm.conformers.cluster(mol, centroids=False)
    assert len(clustered_mol) == 2
    assert clustered_mol[0].GetNumConformers() > 30
    assert clustered_mol[1].GetNumConformers() > 5

    # centroids
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None)
    clustered_mol = dm.conformers.cluster(mol, centroids=True)
    assert clustered_mol.GetNumConformers() == 2

    # no centroids - minimize
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None, minimize_energy=True)
    clustered_mol = dm.conformers.cluster(mol, centroids=False)
    assert len(clustered_mol) == 2
    assert clustered_mol[0].GetNumConformers() > 30
    assert clustered_mol[1].GetNumConformers() > 5

    # centroids - minimize
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, rms_cutoff=None, minimize_energy=True)
    clustered_mol = dm.conformers.cluster(mol, centroids=True)
    assert clustered_mol.GetNumConformers() == 2


def test_get_coords():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=1)

    assert dm.conformers.get_coords(mol).shape == (2, 3)


def test_center_of_mass():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=1)

    # geomtric center
    center = dm.conformers.center_of_mass(mol, use_atoms=False)
    coords = dm.conformers.get_coords(mol)
    np.testing.assert_array_almost_equal(coords.mean(axis=0), center)

    # mass center
    center = dm.conformers.center_of_mass(mol, use_atoms=True)
    assert center.shape == (3,)


def test_translate():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=1)

    coords = dm.conformers.get_coords(mol)
    print(coords)

    dm.conformers.translate(mol, [10, 10, 10])
    new_coords = dm.conformers.get_coords(mol)
    print(new_coords - 10)

    np.testing.assert_array_almost_equal(coords, new_coords - 10, decimal=0)


def test_align_conformers():
    smiles_list = [
        "Nc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1ccn(-c2ccccc2)c(=O)c1F",
        "Cc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1cnn(-c2ccccc2)c(=O)c1",
    ]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]
    mols = [dm.conformers.generate(mol, n_confs=1) for mol in mols]

    # Align
    aligned_mols, scores = dm.conformers.align_conformers(mols)

    # Check
    assert len(scores) == len(mols)
    assert len(aligned_mols) == len(mols)

    for i, (mol, aligned_mol) in enumerate(zip(mols, aligned_mols)):
        p1 = mol.GetConformer().GetPositions()
        p2 = aligned_mol.GetConformer().GetPositions()

        if i == 0:
            # The first molecule is the reference so the positions should remain the same.
            assert np.allclose(p1, p2)
        else:
            assert not np.allclose(p1, p2)


def test_align_conformers_O3A():
    smiles_list = [
        "Nc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1ccn(-c2ccccc2)c(=O)c1F",
        "Cc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1cnn(-c2ccccc2)c(=O)c1",
    ]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]
    mols = [dm.conformers.generate(mol, n_confs=1) for mol in mols]

    # Align
    aligned_mols, scores = dm.conformers.align_conformers(mols, backend="O3A")

    # Check
    assert len(scores) == len(mols)
    assert len(aligned_mols) == len(mols)

    for i, (mol, aligned_mol) in enumerate(zip(mols, aligned_mols)):
        p1 = mol.GetConformer().GetPositions()
        p2 = aligned_mol.GetConformer().GetPositions()

        if i == 0:
            # The first molecule is the reference so the positions should remain the same.
            assert np.allclose(p1, p2)
        else:
            assert not np.allclose(p1, p2)


def test_align_conformers_without_conformer():
    smiles_list = [
        "Nc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1ccn(-c2ccccc2)c(=O)c1F",
        "Cc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1cnn(-c2ccccc2)c(=O)c1",
    ]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]

    with pytest.raises(ValueError):
        dm.conformers.align_conformers(mols)


def test_align_conformers_wrong_backend():
    smiles_list = [
        "Nc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1ccn(-c2ccccc2)c(=O)c1F",
        "Cc1cnn(-c2ccccc2)c(=O)c1Cl",
        "Cc1cnn(-c2ccccc2)c(=O)c1",
    ]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]
    mols = [dm.conformers.generate(mol, n_confs=1) for mol in mols]

    with pytest.raises(ValueError):
        dm.conformers.align_conformers(mols, backend="not_supported")


def test_conformers_minimized_sorted():
    smiles = "CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, n_confs=10, minimize_energy=True)

    confs = list(mol.GetConformers())
    energies = np.array([conf.GetPropsAsDict()["rdkit_UFF_energy"] for conf in confs])

    assert np.all(np.diff(energies) >= 0)


def test_conformers_non_minimized_sorted():
    smiles = "CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol, n_confs=10, minimize_energy=False)

    confs = list(mol.GetConformers())
    energies = np.array([conf.GetPropsAsDict()["rdkit_UFF_energy"] for conf in confs])

    assert np.all(np.diff(energies) >= 0)


def test_keep_conformers_from_indice():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=10)

    mol2 = dm.conformers.keep_conformers(mol, 5)
    conf_ids = [conf.GetId() for conf in mol2.GetConformers()]

    assert mol2.GetNumConformers() == 1
    assert conf_ids == [0]


def test_keep_conformers_from_indice_default_conf():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=10)

    mol2 = dm.conformers.keep_conformers(mol)
    conf_ids = [conf.GetId() for conf in mol2.GetConformers()]

    assert mol2.GetNumConformers() == 1
    assert conf_ids == [0]


def test_keep_conformers_from_indices():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=10)

    mol2 = dm.conformers.keep_conformers(mol, [0, 4, 6])
    conf_ids = [conf.GetId() for conf in mol2.GetConformers()]

    assert mol2.GetNumConformers() == 3
    assert conf_ids == [0, 1, 2]


def test_keep_conformers_from_indices_keep_ids():
    mol = dm.to_mol("CC")
    mol = dm.conformers.generate(mol, n_confs=10)

    mol2 = dm.conformers.keep_conformers(mol, [0, 4, 6], assign_id=False)
    conf_ids = [conf.GetId() for conf in mol2.GetConformers()]

    assert mol2.GetNumConformers() == 3
    assert conf_ids == [0, 4, 6]


def test_conformer_energy():
    mol = dm.to_mol("O=C(C)Oc1ccccc1C(=O)O")

    random.seed(42)
    np.random.seed(42)
    mol1 = dm.conformers.generate(mol, ewindow=7)
    assert np.isclose(mol1.GetNumConformers(), 40, atol=5)
    e1 = mol1.GetConformer(1).GetPropsAsDict()
    assert np.isclose(e1["rdkit_UFF_energy"], 35.640740, atol=1)
    assert np.isclose(e1["rdkit_UFF_delta_energy"], 0.246822, atol=0.1)

    mol2 = dm.conformers.generate(mol, forcefield="MMFF94s", eratio=3)
    assert np.isclose(mol2.GetNumConformers(), 23, atol=5)
    e2 = mol2.GetConformer(2).GetPropsAsDict()
    assert np.isclose(e2["rdkit_MMFF94s_energy"], 38.715689, atol=1)
    assert np.isclose(e2["rdkit_MMFF94s_delta_energy"], 2.220522, atol=1)

    mol3 = dm.conformers.generate(mol, forcefield="MMFF94s_noEstat", minimize_energy=True)
    e3 = mol3.GetConformer(3).GetPropsAsDict()
    assert np.isclose(e3["rdkit_MMFF94s_noEstat_energy"], 38.217380, atol=1)
    assert np.isclose(e3["rdkit_MMFF94s_noEstat_delta_energy"], 0.0, atol=0.1)



from typing import Iterable
from typing import Sequence
from typing import Callable
from typing import Optional
from typing import Any
from typing import Union

import collections.abc
import itertools
import contextlib

import joblib
from joblib import Parallel, delayed

from tqdm.auto import tqdm


class JobRunner:
    def __init__(
        self,
        n_jobs: Optional[int] = -1,
        batch_size: Union[int, str] = "auto",
        prefer: Optional[str] = None,
        progress: bool = False,
        total: Optional[int] = None,
        tqdm_kwargs: Optional[dict] = None,
        **job_kwargs: Any,
    ):
        """
        JobRunner with sequential/parallel regimes. The multiprocessing backend use joblib which
        allows taking advantage of its features, while the progress bar use tqdm

        Args:
            n_jobs: Number of process. Use 0 or None to force sequential.
                Use -1 to use all the available processors. For details see
                https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation
            batch_size: Whether to  batch `inputs_list`. You can specify batch_size when the length
                of `inputs_list` is very large (>100k elements). By default, the auto batching of joblib is used.
            prefer: Choose from ['processes', 'threads'] or None. Default to None.
                Soft hint to choose the default backend if no specific backend
                was selected with the parallel_backend context manager. The
                default process-based backend is 'loky' and the default
                thread-based backend is 'threading'. Ignored if the ``backend``
                parameter is specified.
            progress: whether to display progress bar
            total: The number of elements in the iterator. Only used when `progress` is True.
            tqdm_kwargs: Any additional arguments supported by the `tqdm` progress bar.
            **job_kwargs: Any additional arguments supported by `joblib.Parallel`.

        Example:

        ```python
        import datamol as dm
        runner = dm.JobRunner(n_jobs=4, progress=True, prefer="threads")
        results = runner(lambda x: x**2, [1, 2, 3, 4])
        ```
        """

        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.prefer = prefer
        self.job_kwargs = job_kwargs
        self.job_kwargs.update(n_jobs=self.n_jobs, prefer=self.prefer, batch_size=self.batch_size)
        self.no_progress = not progress
        self.total = total
        self.tqdm_kwargs = tqdm_kwargs or {}

    @property
    def is_sequential(self):
        """Check whether the job is sequential or parallel"""
        return (self.n_jobs is None) or (self.n_jobs in [0, 1])

    @staticmethod
    def wrap_fn(fn: Callable, arg_type: Optional[str] = None, **fn_kwargs):
        """Small wrapper around a callable to properly format it's argument"""

        def _run(args: Any):
            if arg_type == "kwargs":
                fn_kwargs.update(**args)
                return fn(**fn_kwargs)
            elif arg_type == "args":
                return fn(*args, **fn_kwargs)
            return fn(args, **fn_kwargs)

        return _run

    def sequential(
        self,
        callable_fn: Callable,
        data: Iterable[Any],
        arg_type: Optional[str] = None,
        **fn_kwargs,
    ):
        """
        Run job in sequential version

        Args:
            callable_fn (callable): function to call
            data (iterable): input data
            arg_type (str, optional): function argument type ('arg'/None or 'args' or 'kwargs')
            **fn_kwargs (dict, optional): optional keyword argument to pass to the callable funciton
        """
        total_length = JobRunner.get_iterator_length(data)

        if self.total is not None:
            self.tqdm_kwargs["total"] = self.total
        elif "total" not in self.tqdm_kwargs:
            self.tqdm_kwargs["total"] = total_length

        if "disable" not in self.tqdm_kwargs:
            self.tqdm_kwargs["disable"] = self.no_progress

        results = [
            JobRunner.wrap_fn(callable_fn, arg_type, **fn_kwargs)(dt)
            for dt in tqdm(data, **self.tqdm_kwargs)
        ]
        return results

    def parallel(
        self,
        callable_fn: Callable,
        data: Iterable[Any],
        arg_type: Optional[str] = None,
        **fn_kwargs,
    ):
        """
        Run job in parallel

        Args:
            callable_fn (callable): function to call
            data (iterable): input data
            arg_type (str, optional): function argument type ('arg'/None or 'args' or 'kwargs')
            **fn_kwargs (dict, optional): optional keyword argument to pass to the callable funciton
        """

        total_length = JobRunner.get_iterator_length(data)

        if self.total is not None:
            self.tqdm_kwargs["total"] = self.total
        elif "total" not in self.tqdm_kwargs:
            self.tqdm_kwargs["total"] = total_length

        if "disable" not in self.tqdm_kwargs:
            self.tqdm_kwargs["disable"] = self.no_progress

        runner = JobRunner._parallel_helper(**self.job_kwargs)
        results = runner(**self.tqdm_kwargs)(
            delayed(JobRunner.wrap_fn(callable_fn, arg_type, **fn_kwargs))(dt) for dt in data
        )

        return results

    def __call__(self, *args, **kwargs):
        """
        Run job using the n_jobs attribute to determine regime
        """
        if self.is_sequential:
            return self.sequential(*args, **kwargs)
        return self.parallel(*args, **kwargs)

    @staticmethod
    def _parallel_helper(**joblib_args):
        r"""
        Parallel helper function for joblib with tqdm support
        """

        def run(**tq_args):
            def tmp(op_iter):
                with _tqdm_callback(tqdm(**tq_args)):
                    return Parallel(**joblib_args)(op_iter)

            return tmp

        return run

    @staticmethod
    def get_iterator_length(data):
        """Attempt to get the length of an iterator"""
        total_length = None
        try:
            total_length = len(data)
        except TypeError:
            # most likely a generator, ignore
            pass
        return total_length


@contextlib.contextmanager
def _tqdm_callback(pbar):
    """Report tqdm update on job completion"""

    class _CompletionCallBack(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            pbar.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    jlib_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = _CompletionCallBack
    try:
        yield pbar
    finally:
        joblib.parallel.BatchCompletionCallBack = jlib_callback
        pbar.close()


def parallelized(
    fn: Callable,
    inputs_list: Iterable[Any],
    scheduler: str = "processes",
    n_jobs: Optional[int] = -1,
    batch_size: Union[int, str] = "auto",
    progress: bool = False,
    arg_type: str = "arg",
    total: Optional[int] = None,
    tqdm_kwargs: Optional[dict] = None,
    **job_kwargs: Any,
) -> Sequence[Optional[Any]]:
    """Run a function in parallel.

    Args:
        fn: The function to run in parallel.
        inputs_list: List of inputs to pass to `fn`.
        scheduler: Choose between ["processes", "threads"]. Defaults
            to None which uses the default joblib "loky" scheduler.
        n_jobs: Number of process. Use 0 or None to force sequential.
                Use -1 to use all the available processors. For details see
                https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation
        batch_size: Whether to automatically batch `inputs_list`. You should only use it when the length
            of `inputs_list` is very large (>100k elements). The length of `inputs_list` must also be
            defined.
        progress: Display a progress bar. Defaults to False.
        arg_type: One of ["arg", "args", "kwargs]:
            - "arg": the input is passed as an argument: `fn(arg)` (default).
            - "args": the input is passed as a list: `fn(*args)`.
            - "kwargs": the input is passed as a map: `fn(**kwargs)`.
        total: The number of elements in the iterator. Only used when `progress` is True.
        tqdm_kwargs: Any additional arguments supported by the `tqdm` progress bar.
        **job_kwargs: Any additional arguments supported by `joblib.Parallel`.

    Returns:
        The results of the execution as a list.
    """

    runner = JobRunner(
        n_jobs=n_jobs,
        batch_size=batch_size,
        progress=progress,
        prefer=scheduler,
        total=total,
        tqdm_kwargs=tqdm_kwargs,
        **job_kwargs,
    )
    return runner(fn, inputs_list, arg_type=arg_type)


def parallelized_with_batches(
    fn: Callable,
    inputs_list: Iterable[Any],
    batch_size: int,
    scheduler: str = "processes",
    n_jobs: Optional[int] = -1,
    progress: bool = False,
    arg_type: str = "arg",
    total: Optional[int] = None,
    tqdm_kwargs: Optional[dict] = None,
    flatten_results: bool = True,
    joblib_batch_size: Union[int, str] = "auto",
    **job_kwargs: Any,
) -> Sequence[Optional[Any]]:
    """Run a function in parallel using batches.

    Args:
        fn: The function to run in parallel. It must accept a batch of `inputs_list`.
        inputs_list: List of inputs to pass to `fn`.
        batch_size: Batch size on which to run `fn`.
        scheduler: Choose between ["processes", "threads"]. Defaults
            to None which uses the default joblib "loky" scheduler.
        n_jobs: Number of process. Use 0 or None to force sequential.
                Use -1 to use all the available processors. For details see
                https://joblib.readthedocs.io/en/latest/parallel.html#parallel-reference-documentation
        progress: Display a progress bar. Defaults to False.
        arg_type: One of ["arg", "args", "kwargs]:
            - "arg": the input is passed as an argument: `fn(arg)` (default).
            - "args": the input is passed as a list: `fn(*args)`.
            - "kwargs": the input is passed as a map: `fn(**kwargs)`.
        total: The number of elements in the iterator. Only used when `progress` is True.
        tqdm_kwargs: Any additional arguments supported by the `tqdm` progress bar.
        flatten_results: Whether to flatten the results.
        joblib_batch_size: It corresponds to the `batch_size` argument of `dm.parallelized` that
            is forwarded to `joblib.Parallel` under the hood.
        **job_kwargs: Any additional arguments supported by `joblib.Parallel`.

    Returns:
        The results of the execution as a list.
    """

    def _batch_iterator(n: int, iterable: Iterable):
        it = iter(iterable)
        while True:
            chunk_it = itertools.islice(it, n)
            try:
                first_el = next(chunk_it)
            except StopIteration:
                return
            yield list(itertools.chain((first_el,), chunk_it))

    # Compute the total number of batches of possible
    if total is not None:
        n_batches = total // batch_size
        n_batches = max(n_batches, 1)
    elif isinstance(inputs_list, collections.abc.Sized):
        n_batches = len(inputs_list) // batch_size
        n_batches = max(n_batches, 1)
    else:
        n_batches = None

    # Make an iterator over batches so it works even with Iterator without a defined length
    input_chunks = _batch_iterator(batch_size, inputs_list)

    runner = JobRunner(
        n_jobs=n_jobs,
        batch_size=joblib_batch_size,
        progress=progress,
        prefer=scheduler,
        total=n_batches,
        tqdm_kwargs=tqdm_kwargs,
        **job_kwargs,
    )
    results = runner(fn, input_chunks, arg_type=arg_type)

    # Flatten the results
    if flatten_results:
        results = [item for sublist in results for item in sublist]

    return results



import pytest

import datamol as dm
import numpy as np
import pandas as pd


MOLAR_TEST_VALUES = pd.DataFrame(
    [
        (1, 6, "uM"),
        (0.059, 7.229147988357856, "uM"),
        (0.024, 7.61978876, "uM"),
        (0.187, 6.72815839, "uM"),
        (0.00154, 8.8124793, "uM"),
        (128, 6.892790, "nM"),
        (0.000128, 6.892790, "mM"),
    ],
    columns=["xc50", "pxc50", "unit"],
)


def test_molar_to_log():
    # test scalar
    value, log_value, unit = MOLAR_TEST_VALUES.iloc[0].values
    assert dm.molar.molar_to_log(value, unit=unit) == log_value

    # test arrays
    for unit in ["uM", "mM", "nM"]:
        mask = MOLAR_TEST_VALUES["unit"] == unit
        values = MOLAR_TEST_VALUES[mask]["xc50"].tolist()
        log_values = MOLAR_TEST_VALUES[mask]["pxc50"].tolist()
        np.testing.assert_almost_equal(dm.molar.molar_to_log(values, unit=unit), log_values)

    # test wrong unit
    with pytest.raises(ValueError):
        dm.molar.molar_to_log(0.000128, unit="kcal/mol")


def test_log_to_molar():
    # test scalar
    value, log_value, unit = MOLAR_TEST_VALUES.iloc[0].values
    np.testing.assert_almost_equal(dm.molar.log_to_molar(log_value, unit=unit), value)

    # test arrays
    for unit in ["uM", "mM", "nM"]:
        mask = MOLAR_TEST_VALUES["unit"] == unit
        values = MOLAR_TEST_VALUES[mask]["xc50"].tolist()
        log_values = MOLAR_TEST_VALUES[mask]["pxc50"].tolist()
        np.testing.assert_almost_equal(
            dm.molar.log_to_molar(log_values, unit=unit), values, decimal=5
        )

    # test wrong unit
    with pytest.raises(ValueError):
        dm.molar.log_to_molar(7.214, unit="kcal/mol")


def test_log_to_molar_with_integer():
    dm.molar.log_to_molar(6, unit="uM")



"""A set of utility functions to convert between various units and formats used in drug discovery.
"""

from typing import Union
from typing import Iterable

import numpy as np


_MOLAR_SCALES = {"M": 1, "mM": 1e-3, "uM": 1e-6, "nM": 1e-9, "pM": 1e-12, "fM": 1e-15}


def molar_to_log(
    values: Union[float, Iterable[float], np.ndarray],
    unit: str,
) -> Union[float, Iterable[float], np.ndarray]:
    """Convert a molar concentration (XC50 for example) to its log scaled value (pXC50).

    Args:
        values: A molar concentration (can be a scalar, a list or an array).
        unit: The unit of the input concentration. Choose from:
            `{'M', 'fM', 'mM', 'nM', 'pM', 'uM'}`.
    """

    if unit not in _MOLAR_SCALES:
        raise ValueError(
            f"The unit '{unit}' is not supported. Choose from {set(_MOLAR_SCALES.keys())}."
        )

    return -1 * np.log10(np.array(values) * _MOLAR_SCALES[unit])


def log_to_molar(
    values: Union[float, Iterable[float], np.ndarray],
    unit: str,
) -> Union[float, Iterable[float], np.ndarray]:
    """Convert a log-scaled molar concentration (pXC50 for example) to its unscaled value (XC50).

    Args:
        values: A log-scaled molar concentration (can be a scalar, a list or an array).
        unit: The unit of the input concentration. Choose from:
            `{'M', 'fM', 'mM', 'nM', 'pM', 'uM'}`.
    """

    if unit not in _MOLAR_SCALES:
        raise ValueError(
            f"The unit '{unit}' is not supported. Choose from {set(_MOLAR_SCALES.keys())}."
        )

    return 10 ** (-1 * np.array(values, dtype="float")) / _MOLAR_SCALES[unit]



from typing import Union
from typing import List
from typing import Tuple
from typing import Optional
from typing import Any

import fsspec

from rdkit.Chem import Draw

import PIL

import datamol as dm


from .utils import prepare_mol_for_drawing


def to_image(
    mols: Union[List[dm.Mol], dm.Mol],
    legends: Union[List[Union[str, None]], str, None] = None,
    n_cols: int = 4,
    use_svg: bool = True,
    mol_size: Union[Tuple[int, int], int] = (300, 300),
    highlight_atom: Optional[List[List[int]]] = None,
    highlight_bond: Optional[List[List[int]]] = None,
    outfile: Optional[str] = None,
    max_mols: int = 32,
    copy: bool = True,
    indices: bool = False,
    bond_indices: bool = False,
    bond_line_width: int = 2,
    stereo_annotations: bool = True,
    legend_fontsize: int = 16,
    kekulize: bool = True,
    align: Union[dm.Mol, str, bool] = False,
    **kwargs: Any,
):
    """Generate an image out of a molecule or a list of molecules.

    Args:
        mols: One or a list of molecules.
        legends: A string or a list of string as legend for every molecules.
        n_cols: Number of molecules per column.
        use_svg: Whether to ouput an SVG (or a PNG).
        mol_size: A int or a tuple of int defining the size per molecule.
        highlight_atom: the atoms to highlight.
        highlight_bond: The bonds to highlight.
        outfile: Path where to save the image (local or remote path).
        max_mols: The maximum number of molecules to display.
        copy: Whether to copy the molecules or not.
        indices: Whether to draw the atom indices.
        bond_indices: Whether to draw the bond indices.
        bond_line_width: The width of the bond lines.
        legend_fontsize: Font size for the legend.
        kekulize: Run kekulization routine on molecules. Skipped if fails.
        align: Whether to align the 2D coordinates of the molecules.
            - If set to True, align all molecules with `dm.align.auto_align_many()`.
            - If set to a molecule, it is used as a template for alignment with `dm.align.template_align()`.
            - If set to False, no alignment is performed.
            For a more custom alignment, we suggest using directly the module `dm.align` instead.
        **kwargs: Additional arguments to pass to the drawing function. See RDKit
            documentation related to `MolDrawOptions` for more details at
            https://www.rdkit.org/docs/source/rdkit.Chem.Draw.rdMolDraw2D.html.
    """

    if isinstance(mol_size, int):
        mol_size = (mol_size, mol_size)

    if isinstance(mols, dm.Mol):
        mols = [mols]

    if isinstance(legends, str):
        legends = [legends]

    if copy:
        mols = [dm.copy_mol(mol) for mol in mols]

    if max_mols is not None:
        mols = mols[:max_mols]

        if legends is not None:
            legends = legends[:max_mols]

    # Whether to align the molecules
    if isinstance(align, (dm.Mol, str)):
        mols = [dm.align.template_align(mol, template=align) for mol in mols]
    elif align is True:
        mols = dm.align.auto_align_many(mols)

    # Prepare molecules before drawing
    mols = [prepare_mol_for_drawing(mol, kekulize=kekulize) for mol in mols]

    _highlight_atom = highlight_atom
    if highlight_atom is not None and isinstance(highlight_atom[0], int):
        _highlight_atom = [highlight_atom]

    _highlight_bond = highlight_bond
    if highlight_bond is not None and isinstance(highlight_bond[0], int):
        _highlight_bond = [highlight_bond]

    # Don't make the image bigger than it
    if len(mols) < n_cols:
        n_cols = len(mols)

    draw_options = Draw.rdMolDraw2D.MolDrawOptions()
    draw_options.legendFontSize = legend_fontsize
    draw_options.addAtomIndices = indices
    draw_options.addBondIndices = bond_indices
    draw_options.addStereoAnnotation = stereo_annotations
    draw_options.bondLineWidth = bond_line_width

    # Add the custom drawing options.
    _kwargs = {}
    for k, v in kwargs.items():
        if hasattr(draw_options, k):
            setattr(draw_options, k, v)
        else:
            _kwargs[k] = v

    image = Draw.MolsToGridImage(
        mols,
        legends=legends,
        molsPerRow=n_cols,
        useSVG=use_svg,
        subImgSize=mol_size,
        highlightAtomLists=_highlight_atom,
        highlightBondLists=_highlight_bond,
        drawOptions=draw_options,
        **_kwargs,
    )

    if outfile is not None:
        with fsspec.open(outfile, "wb") as f:
            if use_svg:
                if isinstance(image, str):
                    # in a terminal process
                    f.write(image.encode())  # type: ignore
                else:
                    # in a jupyter kernel process
                    f.write(image.data.encode())  # type: ignore
            else:
                if isinstance(image, PIL.PngImagePlugin.PngImageFile):  # type: ignore
                    # in a terminal process
                    image.save(f)
                else:
                    # in a jupyter kernel process
                    f.write(image.data)  # type: ignore

    return image



from ._version import __version__

from ._version import is_lower_than_current_rdkit_version
from ._version import is_greater_than_current_rdkit_version
from ._version import is_lower_eq_than_current_rdkit_version
from ._version import is_greater_eq_than_current_rdkit_version

from .types import Mol
from .types import BondType
from .types import ChemicalReaction
from .types import Atom
from .types import Bond

from . import utils

from .utils import parallelized
from .utils import parallelized_with_batches
from .utils import JobRunner
from .utils import fs

from .data import freesolv
from .data import cdk2
from .data import solubility

from .log import enable_rdkit_log
from .log import disable_rdkit_log
from .log import without_rdkit_log

from .mol import PERIODIC_TABLE
from .mol import TRIPLE_BOND
from .mol import DOUBLE_BOND
from .mol import SINGLE_BOND
from .mol import AROMATIC_BOND
from .mol import UNSPECIFIED_BOND

from .mol import copy_mol
from .mol import to_mol
from .mol import same_mol
from .mol import reorder_atoms
from .mol import randomize_atoms
from .mol import to_neutral
from .mol import sanitize_mol
from .mol import sanitize_first
from .mol import sanitize_smiles
from .mol import standardize_smiles
from .mol import standardize_mol
from .mol import fix_valence_charge
from .mol import incorrect_valence
from .mol import decrease_bond
from .mol import fix_valence
from .mol import adjust_singleton
from .mol import remove_dummies
from .mol import fix_mol
from .mol import replace_dummies_atoms
from .mol import keep_largest_fragment
from .mol import is_transition_metal
from .mol import set_dative_bonds
from .mol import set_mol_props
from .mol import copy_mol_props
from .mol import atom_indices_to_mol
from .mol import protect_atoms
from .mol import atom_list_to_bond
from .mol import substructure_matching_bonds
from .mol import add_hs
from .mol import remove_hs
from .mol import unique_id
from .mol import hash_mol
from .mol import clear_mol_props
from .mol import strip_mol_to_core
from .mol import make_scaffold_generic
from .mol import to_scaffold_murcko
from .mol import compute_ring_system
from .mol import clear_atom_props
from .mol import clear_atom_map_number
from .mol import set_atom_positions
from .mol import get_atom_positions

from .cluster import cluster_mols
from .cluster import pick_diverse
from .cluster import pick_centroids
from .cluster import assign_to_centroids

from . import fragment
from . import scaffold
from . import molar
from . import descriptors
from . import predictors
from . import reactions

from .convert import to_smiles
from .convert import to_selfies
from .convert import from_selfies
from .convert import to_smarts
from .convert import from_smarts
from .convert import smiles_as_smarts
from .convert import to_inchi
from .convert import to_inchikey
from .convert import from_inchi
from .convert import to_df
from .convert import from_df
from .convert import render_mol_df
from .convert import to_inchi_non_standard
from .convert import to_inchikey_non_standard

from .fp import to_fp
from .fp import fp_to_array
from .fp import list_supported_fingerprints
from .fp import fold_count_fp

from .similarity import pdist
from .similarity import cdist

from .graph import to_graph
from .graph import get_all_path_between

from .io import read_csv
from .io import read_excel
from .io import read_sdf
from .io import to_sdf
from .io import to_smi
from .io import read_smi
from .io import read_mol2file
from .io import read_molblock
from .io import to_molblock
from .io import to_xlsx
from .io import read_pdbblock
from .io import to_pdbblock
from .io import read_pdbfile
from .io import to_pdbfile

from .isomers import enumerate_stereoisomers
from .isomers import enumerate_tautomers
from .isomers import enumerate_structisomers
from .isomers import canonical_tautomer
from .isomers import remove_stereochemistry

from . import align
from . import conformers
from . import viz

from .viz import to_image

from .mcs import find_mcs

from .graph import to_graph
from .graph import get_all_path_between
from .graph import match_molecular_graphs
from .graph import reorder_mol_from_template



from typing import Callable
from typing import Dict
from typing import List
from typing import Union
from typing import Optional

import functools

import pandas as pd

from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors

from .. import Mol
from ..utils.jobs import parallelized

from .descriptors import mw
from .descriptors import fsp3
from .descriptors import n_lipinski_hba
from .descriptors import n_lipinski_hbd
from .descriptors import n_rings
from .descriptors import n_hetero_atoms
from .descriptors import n_heavy_atoms
from .descriptors import n_rotatable_bonds
from .descriptors import n_radical_electrons
from .descriptors import tpsa
from .descriptors import qed
from .descriptors import clogp
from .descriptors import sas
from .descriptors import n_aliphatic_carbocycles
from .descriptors import n_aliphatic_heterocyles
from .descriptors import n_aliphatic_rings
from .descriptors import n_aromatic_carbocycles
from .descriptors import n_aromatic_heterocyles
from .descriptors import n_aromatic_rings
from .descriptors import n_saturated_carbocycles
from .descriptors import n_saturated_heterocyles
from .descriptors import n_saturated_rings


def any_rdkit_descriptor(name: str) -> Callable:
    """Return a descriptor function by name either from
    `rdkit.Chem import Descriptors` or `rdkit.Chem.rdMolDescriptors`.

    Args:
        name: Descriptor name.
    """
    fn = getattr(Descriptors, name, None)

    if fn is None:
        fn = getattr(rdMolDescriptors, name, None)

    if fn is None:
        raise ValueError(f"Descriptor {name} not found.")

    return fn


_DEFAULT_PROPERTIES_FN = {
    "mw": mw,
    "fsp3": fsp3,
    "n_lipinski_hba": n_lipinski_hba,
    "n_lipinski_hbd": n_lipinski_hbd,
    "n_rings": n_rings,
    "n_hetero_atoms": n_hetero_atoms,
    "n_heavy_atoms": n_heavy_atoms,
    "n_rotatable_bonds": n_rotatable_bonds,
    "n_radical_electrons": n_radical_electrons,
    "tpsa": tpsa,
    "qed": qed,
    "clogp": clogp,
    "sas": sas,
    "n_aliphatic_carbocycles": n_aliphatic_carbocycles,
    "n_aliphatic_heterocyles": n_aliphatic_heterocyles,
    "n_aliphatic_rings": n_aliphatic_rings,
    "n_aromatic_carbocycles": n_aromatic_carbocycles,
    "n_aromatic_heterocyles": n_aromatic_heterocyles,
    "n_aromatic_rings": n_aromatic_rings,
    "n_saturated_carbocycles": n_saturated_carbocycles,
    "n_saturated_heterocyles": n_saturated_heterocyles,
    "n_saturated_rings": n_saturated_rings,
}


def compute_many_descriptors(
    mol: Mol,
    properties_fn: Optional[Dict[str, Union[Callable, str]]] = None,
    add_properties: bool = True,
) -> dict:
    """Compute a list of opiniated molecular properties.

    Args:
        mol: A molecule.
        properties_fn: A list of functions that compute properties. If None,
            a default list of properties is used. If the function is a string,
            `dm.descriptors.any_descriptor()` is used to retrieve the descriptor
            function.
        add_properties: Whether to add the computed properties to the default list.

    Returns:
        Computed properties as a dict.
    """

    if properties_fn is None:
        properties_fn = _DEFAULT_PROPERTIES_FN
    elif add_properties:
        [properties_fn.setdefault(k, v) for k, v in _DEFAULT_PROPERTIES_FN.items()]

    props = {}
    for k, v in properties_fn.items():
        if isinstance(v, str):
            v = any_rdkit_descriptor(v)

        props[k] = v(mol)

    return props


def batch_compute_many_descriptors(
    mols: List[Mol],
    properties_fn: Optional[Dict[str, Union[Callable, str]]] = None,
    add_properties: bool = True,
    n_jobs: int = 1,
    batch_size: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = True,
) -> pd.DataFrame:
    """Compute a list of opiniated molecular properties on a list of molecules.

    Args:
        mols: A list of molecules.
        properties_fn: A list of functions that compute properties. If None,
            a default list of properties is used. If the function is a string,
            `dm.descriptors.any_descriptor()` is used to retrieve the descriptor
            function.
        add_properties: Whether to add the computed properties to the default list.

    Returns:
        A dataframe of computed properties with one row per input molecules.
    """

    compute_fn = functools.partial(
        compute_many_descriptors,
        properties_fn=properties_fn,
        add_properties=add_properties,
    )

    props = parallelized(
        compute_fn,
        mols,
        batch_size=batch_size,
        progress=progress,
        n_jobs=n_jobs,
        tqdm_kwargs=dict(leave=progress_leave),
    )
    return pd.DataFrame(props)



from typing import Optional

import time
import signal

from loguru import logger


duration_intervals = (
    ("weeks", 604800),  # 60 * 60 * 24 * 7
    ("days", 86400),  # 60 * 60 * 24
    ("h", 3600),  # 60 * 60
    ("min", 60),
    ("s", 1),
    ("ms", 1e-3),
    ("us", 1e-6),
)


def human_duration(seconds: float, granularity: int = 1):
    # NOTE(hadim): far from being perfect.

    result = []
    duration: float = seconds
    for name, count in duration_intervals:
        value = duration // count
        if value:
            duration -= value * count
            result.append(f"{value:.0f}{name}")
    return ", ".join(result[:granularity])


class watch_duration:
    """A Python decorator to measure execution time with logging capability.

    Args:
        log: Whether to log the measured duration.
        log_human_duration: Whether to log duration in a human way
            depending on the amount.

    Example:

    ```python
    def fn(n):
        for i in range(n):
            print(i)
            time.sleep(0.2)

    with dm.utils.perf.watch_duration(log=True) as w:
        fn(5)

    print(w.duration)
    ```
    """

    def __init__(self, log: bool = True, log_human_duration: bool = True):
        self.log = log
        self.log_human_duration = log_human_duration

        self.start = None
        self.end = None
        self.duration = None
        self.duration_minutes = None

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *_):
        assert self.start is not None

        self.end = time.time()
        self.duration = self.end - self.start
        self.duration_minutes = self.duration / 60

        if self.log:
            if self.log_human_duration:
                logger.info(f"Duration {human_duration(self.duration)}.")
            else:
                logger.info(f"Duration {self.duration_minutes:.2f} minutes")



import platform
import pathlib
from loguru import logger

import pytest


DATA_DIR_PATH = pathlib.Path(__file__).parent.resolve() / "data"


@pytest.fixture
def current_platform():
    if platform.system() == "Linux":
        return "linux"
    elif platform.system() == "Darwin":
        return "osx"
    elif platform.system() == "Windows":
        return "win"
    else:
        return platform.system()


@pytest.fixture(autouse=True)
def skip_by_platform(request, current_platform):
    if request.node.get_closest_marker("skip_platform"):
        if request.node.get_closest_marker("skip_platform").args[0] == current_platform:
            pytest.skip(f"skipped on this platform: {current_platform}")


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "skip_platform(current_platform): skip test for a given platform from `['linux', 'osx', 'win']`",
    )


@pytest.fixture
def datadir(request):
    return DATA_DIR_PATH


from _pytest.logging import caplog as _caplog  # Mandatory for the below monkeypatch function.


@pytest.fixture
def caplog(_caplog):
    """Monkeypatching the pytest caplog to work with loguru.

    See https://loguru.readthedocs.io/en/latest/resources/migration.html#making-things-work-with-pytest-and-caplog
    """
    import logging

    class PropogateHandler(logging.Handler):
        def emit(self, record):
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(PropogateHandler(), format="{message}")
    yield _caplog
    logger.remove(handler_id)



from typing import Callable
from typing import List
from typing import Union

import platform
from functools import wraps


def disable_on_os(os_names: Union[str, List[str]]):
    """A decorator to disable a function raising an error if the OS detected is not supported.

    Args:
        os_names: OS names to disable this function. Valid OS names are: `["linux", "osx", "win"]`.
    """

    if isinstance(os_names, str):
        os_names = [os_names]

    valid_os_names = []
    for os_name in os_names:
        if os_name == "linux":
            valid_os_names.append("Linux")
        elif os_name == "win":
            valid_os_names.append("Windows")
        elif os_name == "osx":
            valid_os_names.append("Darwin")
        else:
            valid_os_names.append(os_name)

    def real_decorator(function: Callable):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if platform.system() not in valid_os_names:
                retval = function(*args, **kwargs)
                return retval
            else:
                raise NotImplementedError(
                    f"The function {function.__name__} is not supported"
                    f" for the platform '{platform.system()}'."
                )

        return wrapper

    return real_decorator



import pytest

import base64
import io

import numpy as np
import ipywidgets as widgets

import PIL
from PIL import Image

import datamol as dm


# NOTE(hadim): rdkit returns different image objects
# according to the Python process context (Jupyter notebook vs terminal).
# In consequence, those tests will fail if they are executed within a
# Jupyter notebook.


def _convert_ipython_to_array(image):
    """convert ipython image to numpy array"""
    image_obj = base64.b64decode(str(image._repr_png_()))
    try:
        image_obj = Image.open(io.BytesIO(image_obj))
        return np.array(image_obj)
    except:
        return np.array(image)


def test_to_image():
    # Get a list of molecules
    data = dm.data.freesolv()
    mols = dm.from_df(data)  # type: ignore
    mols = mols[:8]

    # With multiple molecules
    legends = [dm.to_smiles(mol) for mol in mols]
    image = dm.viz.to_image(mols, legends=legends, n_cols=4, mol_size=(200, 200), use_svg=False)
    image = _convert_ipython_to_array(image)

    print(type(image))

    image = np.array(image)

    assert image.dtype == np.uint8
    assert image.shape == (400, 800, 3)
    assert image.shape[1] == 200 * 4

    # With a single molecule
    mol = mols[0]
    legends = dm.to_smiles(mol)
    image = dm.viz.to_image(mol, legends=legends, mol_size=(200, 200), use_svg=False)
    image = _convert_ipython_to_array(image)
    image = np.array(image)

    assert image.dtype == np.uint8
    assert image.shape == (200, 200, 3)

    dm.viz.to_image(mol, indices=True, mol_size=400)


def test_to_image_incorrect_aromaticity():
    query = "C-c1cn(-C-2-[N,O:3]-[#6@H](-C-[#6,#8:1]-[*:2])-C(-[#8])-C-2-[#1,#8,#9:4])c2ncnc(-C)c12"
    mol = dm.from_smarts(query)
    dm.to_image(
        mol,
        mol_size=300,
        use_svg=False,
        legends="a legend",
        legend_fontsize=40,
        stereo_annotations=False,
    )


def test_to_image_save_file(tmpdir):
    smiles = "CCCOCc1cc(c2ncccc2)ccc1"
    mol = dm.to_mol(smiles)

    image_path = str(tmpdir.join("mol.png"))
    dm.viz.to_image(mol, outfile=image_path, use_svg=False)

    # check whether the png is valid
    try:
        img = Image.open(image_path)
        img.verify()
    except PIL.UnidentifiedImageError:
        pytest.fail(f"The image {image_path} is invalid.")

    image_path = str(tmpdir.join("mol.svg"))
    dm.viz.to_image(mol, outfile=image_path, use_svg=True)

    # check whether the svg looks valid
    with open(image_path) as f:
        content = f.read().strip()
    assert content.startswith("<?xml ") or content.startswith("<svg")
    assert content.endswith("</svg>")


def test_conformers():
    import nglview as nv

    smiles = "CCCC=O"
    mol = dm.to_mol(smiles)
    mol = dm.conformers.generate(mol)

    # one conformer
    view = dm.viz.conformers(mol)
    assert type(view) == nv.widget.NGLWidget

    # multiple conformers
    view = dm.viz.conformers(mol, n_confs=12)
    assert type(view) == widgets.GridspecLayout


def test_circle_grid(tmp_path):
    mol = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")

    im = dm.viz.circle_grid(
        mol,
        [
            [dm.to_mol("CCC"), dm.to_mol("CCCCCCC")],
            [dm.to_mol("CCCO"), dm.to_mol("CCCCCCCO")],
        ],
    )

    im.save(tmp_path / "image.png")


def test_to_image_align():
    # Get a list of molecules
    data = dm.data.freesolv()
    mols = dm.from_df(data)  # type: ignore
    mols = mols[:8]

    # With multiple molecules
    dm.viz.to_image(mols, align=True)


def test_to_image_align_template():
    # Get a list of molecules
    data = dm.data.freesolv()
    mols = dm.from_df(data)  # type: ignore
    mols = mols[:8]

    dm.viz.to_image(mols, align=mols[0])



import sys
import os

from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import RDConfig
from rdkit.Chem import Lipinski
from rdkit.Chem import rdmolops
from rdkit.Chem import Crippen


from .. import Mol
from ..convert import from_smarts
from .._version import is_lower_than_current_rdkit_version


def _sasscorer(mol: Mol):
    sys.path.append(os.path.join(RDConfig.RDContribDir, "SA_Score"))
    try:
        import sascorer  # type:ignore
    except ImportError:
        raise ImportError(
            "Could not import sascorer. If you installed rdkit-pypi with `pip`, please uninstall it and reinstall rdkit with `conda` or `mamba`."
        )

    return sascorer.calculateScore(mol)


_AROMATIC_QUERY = from_smarts("a")

mw = rdMolDescriptors.CalcExactMolWt
fsp3 = rdMolDescriptors.CalcFractionCSP3
tpsa = rdMolDescriptors.CalcTPSA
qed = Descriptors.qed
clogp = Descriptors.MolLogP  # type: ignore
sas = _sasscorer
formal_charge = rdmolops.GetFormalCharge
refractivity = Crippen.MolMR

n_hba = rdMolDescriptors.CalcNumHBA
n_hbd = rdMolDescriptors.CalcNumHBD
n_lipinski_hba = rdMolDescriptors.CalcNumLipinskiHBA
n_lipinski_hbd = rdMolDescriptors.CalcNumLipinskiHBD
n_rings = rdMolDescriptors.CalcNumRings
n_hetero_atoms = rdMolDescriptors.CalcNumHeteroatoms


if is_lower_than_current_rdkit_version("2021.09"):
    n_heavy_atoms = Descriptors.HeavyAtomCount  # type: ignore
else:
    n_heavy_atoms = rdMolDescriptors.CalcNumHeavyAtoms

n_rotatable_bonds = rdMolDescriptors.CalcNumRotatableBonds
n_radical_electrons = Descriptors.NumRadicalElectrons
n_NHOH = Lipinski.NHOHCount
n_NO = Lipinski.NOCount
n_spiro_atoms = rdMolDescriptors.CalcNumSpiroAtoms

n_aliphatic_carbocycles = rdMolDescriptors.CalcNumAliphaticCarbocycles
n_aliphatic_heterocyles = rdMolDescriptors.CalcNumAliphaticHeterocycles
n_aliphatic_rings = rdMolDescriptors.CalcNumAliphaticRings

n_aromatic_carbocycles = rdMolDescriptors.CalcNumAromaticCarbocycles
n_aromatic_heterocyles = rdMolDescriptors.CalcNumAromaticHeterocycles
n_aromatic_rings = rdMolDescriptors.CalcNumAromaticRings

n_saturated_carbocycles = rdMolDescriptors.CalcNumSaturatedCarbocycles
n_saturated_heterocyles = rdMolDescriptors.CalcNumSaturatedHeterocycles
n_saturated_rings = rdMolDescriptors.CalcNumSaturatedRings


def n_rigid_bonds(mol: Mol) -> int:
    """Compute the number of rigid bonds in a molecule.

    Rigid bonds are bonds that are not single and not in rings.

    Args:
        mol: A molecule.

    Returns:
        n_rigid_bonds: number of rigid bonds in the molecule
    """
    non_rigid_bonds_count = from_smarts("*-&!@*")
    n_rigid_bonds = mol.GetNumBonds() - len(mol.GetSubstructMatches(non_rigid_bonds_count))
    return n_rigid_bonds


def n_aromatic_atoms(mol: Mol) -> int:
    """Calculate the number of aromatic atoms."""
    matches = mol.GetSubstructMatches(_AROMATIC_QUERY)
    return len(matches)


def n_aromatic_atoms_proportion(mol: Mol) -> int:
    """Calculate the aromatic proportion: # aromatic atoms/#atoms total.

    Args:
        mol: A molecule.

    Only heavy atoms are considered.
    """
    return n_aromatic_atoms(mol) / mol.GetNumHeavyAtoms()


def n_stereo_centers(mol: Mol) -> int:
    """Compute the number of stereocenters in a molecule.

    Args:
        mol: A molecule.

    Returns:
        n_stero_center: number of stereocenters in the molecule
    """
    n = 0
    try:
        rdmolops.FindPotentialStereo(mol, cleanIt=False)
        n = rdMolDescriptors.CalcNumAtomStereoCenters(mol)
    except:
        pass
    return n


def n_stereo_centers_unspecified(mol: Mol) -> int:
    """Compute the number of unspecified stereocenters in a molecule.

    Args:
        mol: A molecule.

    Returns:
        n_stereo_centers_unspecified: number of unspecified stereocenters in the molecule
    """
    n = 0
    try:
        rdmolops.FindPotentialStereo(mol, cleanIt=False)
        n = rdMolDescriptors.CalcNumUnspecifiedAtomStereoCenters(mol)
    except:
        pass
    return n


def n_charged_atoms(mol: Mol) -> int:
    """Compute the number of charged atoms in a molecule.

    Args:
        mol: A molecule.

    Returns:
        n_charged_atoms: number of charged atoms in the molecule
    """
    return sum([at.GetFormalCharge() != 0 for at in mol.GetAtoms()])



import unittest

import numpy as np

import datamol as dm


def test_cluster_mols():
    # Get some mols
    data = dm.data.freesolv()
    smiles = data["smiles"].iloc[:100].tolist()
    mols = [dm.to_mol(s) for s in smiles]

    _, mol_clusters = dm.cluster_mols(mols, cutoff=0.7)
    cluster_sizes = [11, 7, 5, 3, 3, 3, 2, 3, 2, 1, 2, 2, 1]
    assert [len(c) for c in mol_clusters[:13]] == cluster_sizes


def test_pick_diverse():
    # Get some mols
    data = dm.data.freesolv()
    smiles = data["smiles"].iloc[:100].tolist()
    mols = [dm.to_mol(s) for s in smiles]

    indices, _ = dm.pick_diverse(mols, npick=18, seed=19)

    excepted_indices = np.array(
        [9, 14, 47, 50, 56, 61, 67, 89, 83, 90, 94, 10, 0, 96, 15, 58, 71, 21]
    )

    assert np.all(indices == excepted_indices)


def test_pick_centroids():
    data = dm.data.freesolv()
    smiles = data["smiles"].iloc[:100].tolist()
    mols = [dm.to_mol(s) for s in smiles]
    indices, centroids = dm.pick_centroids(
        mols, npick=18, threshold=0.7, method="sphere", n_jobs=-1
    )
    excepted_indices = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 13, 14, 15, 16, 17, 18, 20])

    assert np.all(indices == excepted_indices)


def test_assign_to_centroids():
    data = dm.data.freesolv()
    smiles = data["smiles"].iloc[:100].tolist()
    mols = [dm.to_mol(s) for s in smiles]
    indices, centroids = dm.pick_centroids(
        mols, npick=18, threshold=0.7, method="sphere", n_jobs=-1
    )

    cluster_map, cluster_list = dm.assign_to_centroids(mols, centroids, n_jobs=-1)
    # expect centroid to be in centroid list
    assert indices[0] in cluster_map[0]
    # expect no intersection after assignment
    map_intersection = set.intersection(*map(set, cluster_map.values()))
    assert len(map_intersection) == 0
    # expect some similar molecule in a given cluster
    # assert 33 in cluster_map[0]



from typing import List
from typing import Optional
from typing import Union

import functools

import numpy as np
from scipy.spatial import distance

from rdkit import Chem
from rdkit.DataManip.Metric import GetTanimotoDistMat  # type: ignore
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity

import datamol as dm


def pdist_rdkit(
    mols: List[Union[str, Chem.rdchem.Mol]],
    n_jobs: Optional[int] = 1,
    squareform: bool = True,
    **fp_args,
) -> np.ndarray:
    """Equivalent to `dm.similarity.pdist` but uses the RDKit API.

    Important:
        This function is only used for testing and shoult not be used in production.
    """

    fps = dm.parallelized(
        functools.partial(dm.to_fp, as_array=False, **fp_args),
        mols,
        n_jobs=n_jobs,
    )

    fps = list(fps)  # type: ignore

    dist = GetTanimotoDistMat(fps)

    # Put in squareform: `scipy.spatial.distance.squareform` is incompatible with RDKit returned vector.
    dist_mat = np.zeros((len(fps), len(fps)))
    dist_mat[np.tril_indices_from(dist_mat, -1)] = dist
    dist_mat += dist_mat.T

    if not squareform:
        dist_mat = distance.squareform(dist_mat, force="tovector")

    return dist_mat


def cdist_rdkit(
    mols1: List[Union[str, Chem.rdchem.Mol]],
    mols2: List[Union[str, Chem.rdchem.Mol]],
    n_jobs: Optional[int] = 1,
    **fp_args,
) -> np.ndarray:
    """Equivalent to `dm.similarity.cdist` but uses the RDKit API.

    Important:
        This function is only used for testing and shoult not be used in production.
    """

    fps1 = dm.parallelized(
        functools.partial(dm.to_fp, as_array=False, **fp_args),
        mols1,
        n_jobs=n_jobs,
    )

    fps2 = dm.parallelized(
        functools.partial(dm.to_fp, as_array=False, **fp_args),
        mols2,
        n_jobs=n_jobs,
    )

    fps1 = list(fps1)  # type: ignore
    fps2 = list(fps2)  # type: ignore

    dist_mat = np.zeros((len(fps1), len(fps2)))
    for i in range(len(fps1)):
        for j in range(len(fps2)):
            d = 1 - TanimotoSimilarity(fps1[i], fps2[j])
            dist_mat[i, j] = d

    return dist_mat



from typing import Union
from typing import Optional
from typing import Any

import numpy as np

import datamol as dm

from rdkit import Chem
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import rdmolops
from rdkit.Chem import rdReducedGraphs
from rdkit.Chem.EState import Fingerprinter as EStateFingerprinter
from rdkit.Avalon import pyAvalonTools

from rdkit.DataStructs.cDataStructs import SparseBitVect
from rdkit.DataStructs.cDataStructs import UIntSparseIntVect
from rdkit.DataStructs.cDataStructs import IntSparseIntVect
from rdkit.DataStructs.cDataStructs import LongSparseIntVect
from rdkit.DataStructs.cDataStructs import ExplicitBitVect
from rdkit.DataStructs.cDataStructs import ULongSparseIntVect


_FP_FUNCS = {
    "maccs": rdMolDescriptors.GetMACCSKeysFingerprint,
    "ecfp": rdMolDescriptors.GetMorganFingerprintAsBitVect,
    "fcfp": rdMolDescriptors.GetMorganFingerprintAsBitVect,
    "topological": rdMolDescriptors.GetHashedTopologicalTorsionFingerprintAsBitVect,
    "atompair": rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect,
    "rdkit": rdmolops.RDKFingerprint,
    "pattern": rdmolops.PatternFingerprint,
    "layered": rdmolops.LayeredFingerprint,
    "erg": rdReducedGraphs.GetErGFingerprint,
    # NOTE(hadim): bad for pickling?
    "estate": lambda x, **args: EStateFingerprinter.FingerprintMol(x)[0],
    "avalon-count": pyAvalonTools.GetAvalonCountFP,
    "rdkit-count": rdmolops.UnfoldedRDKFingerprintCountBased,
    "ecfp-count": rdMolDescriptors.GetHashedMorganFingerprint,
    "fcfp-count": rdMolDescriptors.GetHashedMorganFingerprint,
    "topological-count": rdMolDescriptors.GetHashedTopologicalTorsionFingerprint,
    "atompair-count": rdMolDescriptors.GetHashedAtomPairFingerprint,
}

_FP_DEFAULT_ARGS = {
    "maccs": {},
    "avalon": {
        "nBits": 512,
        "isQuery": False,
        "resetVect": False,
        "bitFlags": pyAvalonTools.avalonSimilarityBits,
    },
    "ecfp": {
        "radius": 3,  # ECFP6 - not the RDKit default (ECFP4)
        "nBits": 2048,
        "invariants": [],
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": False,
    },
    "fcfp": {
        "radius": 2,  # FCFP4
        "nBits": 2048,
        "invariants": [],  # you may want to provide features invariance
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": True,
    },
    "topological": {
        "nBits": 2048,
        "targetSize": 4,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "nBitsPerEntry": 4,
        "includeChirality": False,
    },
    "atompair": {
        "nBits": 2048,
        "minLength": 1,
        "maxLength": 30,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "nBitsPerEntry": 4,
        "includeChirality": False,
        "use2D": True,
        "confId": -1,
    },
    "rdkit": {
        "minPath": 1,
        "maxPath": 7,
        "fpSize": 2048,
        "nBitsPerHash": 2,
        "useHs": True,
        "tgtDensity": 0.0,
        "minSize": 128,
        "branchedPaths": True,
        "useBondOrder": True,
        "atomInvariants": 0,
        "fromAtoms": 0,
        "atomBits": None,
        "bitInfo": None,
    },
    "pattern": {"fpSize": 2048, "atomCounts": [], "setOnlyBits": None},
    "layered": {
        "fpSize": 2048,
        "minPath": 1,
        "maxPath": 7,
        "atomCounts": [],
        "setOnlyBits": None,
        "branchedPaths": True,
        "fromAtoms": 0,
    },
    "secfp": {
        "n_permutations": 128,
        "nBits": 2048,
        "radius": 3,
        "min_radius": 1,
        "rings": True,
        "kekulize": False,
        "isomeric": False,
        "seed": 0,
    },
    "erg": {"atomTypes": 0, "fuzzIncrement": 0.3, "minPath": 1, "maxPath": 15},
    "estate": {},
    # COUNTING FP
    "ecfp-count": {
        "radius": 2,  # ECFP4
        "nBits": 2048,
        "invariants": [],
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": False,
        "includeRedundantEnvironments": False,
    },
    "fcfp-count": {
        "radius": 2,  # FCFP4
        "nBits": 2048,
        "invariants": [],  # you may want to provide features invariance
        "fromAtoms": [],
        "useChirality": False,
        "useBondTypes": True,
        "useFeatures": True,
        "includeRedundantEnvironments": False,
    },
    "topological-count": {
        "nBits": 2048,
        "targetSize": 4,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "includeChirality": False,
    },
    "avalon-count": {
        "nBits": 512,
        "isQuery": False,
        "bitFlags": pyAvalonTools.avalonSimilarityBits,
    },
    "rdkit-count": {
        "minPath": 1,
        "maxPath": 7,
        "useHs": True,
        "branchedPaths": True,
        "useBondOrder": True,
        "atomInvariants": 0,
        "fromAtoms": 0,
        "atomBits": None,
        "bitInfo": None,
    },
    "atompair-count": {
        "nBits": 2048,
        "minLength": 1,
        "maxLength": 30,
        "fromAtoms": 0,
        "ignoreAtoms": 0,
        "atomInvariants": 0,
        "includeChirality": False,
        "use2D": True,
        "confId": -1,
    },
}


def fp_to_array(
    fp: Union[np.ndarray, SparseBitVect, ExplicitBitVect, UIntSparseIntVect]
) -> np.ndarray:
    """Convert rdkit fingerprint to numpy array.

    Note:
        This implementation has shown to be faster than using `DataStructs.ConvertToNumpyArray`
        by a factor of ~4. See https://github.com/rdkit/rdkit/discussions/3863.

    Args:
        fp: The fingerprint.
    """

    if isinstance(fp, np.ndarray):
        fp_out = fp

    elif isinstance(fp, SparseBitVect):
        tmp = np.zeros(fp.GetNumBits(), dtype=int)
        on_bits = np.array(fp.GetOnBits())
        tmp[on_bits] = 1
        fp_out = tmp

    elif isinstance(fp, ExplicitBitVect):
        fp_out = np.frombuffer(fp.ToBitString().encode(), "u1") - ord("0")

    elif isinstance(
        fp,
        (
            UIntSparseIntVect,
            IntSparseIntVect,
            LongSparseIntVect,
            ULongSparseIntVect,
        ),
    ):
        tmp = np.zeros(fp.GetLength(), dtype=int)
        bit_idx, values = np.array(list(fp.GetNonzeroElements().items())).T
        tmp[bit_idx] = values
        fp_out = tmp

    else:
        raise ValueError(
            f"The fingerprint of type '{type(fp)}' is not supported. "
            "Please open a ticket at https://github.com/datamol-io/datamol/issues."
        )

    return fp_out


def to_fp(
    mol: Union[str, Chem.rdchem.Mol],
    as_array: bool = True,
    fp_type: str = "ecfp",
    fold_size: Optional[int] = None,
    **fp_args: Any,
) -> Optional[Union[np.ndarray, SparseBitVect, ExplicitBitVect]]:
    """Compute the molecular fingerprint given a molecule or a SMILES.

    Args:
        mol: a molecule or a SMILES.
        as_array: Whether to return a numpy array of an RDKit vec. Default to True.
        fp_type: The type of the fingerprint. See `dm.list_supported_fingerprints()` for a
            complete list.
        fold_size: If set, fold the fingerprint to the `fold_size`. If set, returned array is
            always a numpy array.
        **fp_args: Arguments to build the fingerprint. Refer to the official RDKit documentation.

    Returns:
        A fingerprint vector or None
    """

    # Get fp function
    fp_func = _FP_FUNCS.get(fp_type)

    if fp_func is None:
        raise ValueError(
            f"The fingerprint '{fp_type}' is not available. Use `dm.list_supported_fingerprints()` to "
            "get a complete list of the available fingerprints."
        )

    # Convert input to mol if needed
    if isinstance(mol, str):
        mol_obj = dm.to_mol(mol)
    else:
        mol_obj = mol

    if mol_obj is None:
        raise ValueError(f"It seems like the input molecule '{mol}' is invalid.")

    mol = mol_obj

    # Insert default values.
    for key, value in _FP_DEFAULT_ARGS[fp_type].items():
        fp_args.setdefault(key, value)

    # Compute the fingerprint
    fp = fp_func(mol, **fp_args)

    # Fold the fp if needed.
    if fold_size is not None:
        fp = fold_count_fp(fp, dim=fold_size)

    # Convert to a numpy array
    if not fold_size and as_array:
        fp = fp_to_array(fp)

    return fp


def list_supported_fingerprints():
    """Return the supported fingerprints in datamol."""

    return _FP_FUNCS


def fold_count_fp(
    fp: Union[np.ndarray, SparseBitVect, ExplicitBitVect],
    dim: int = 1024,
    binary: bool = False,
) -> np.ndarray:
    """Fast folding of a count fingerprint to the specified dimension.

    Args:
        fp: A fingerprint.
        dim: The dimension of the folded array.
        binary: Whether to fold into a binary array or take use a count vector.

    Returns:
        folded: returns folded array to the provided dimension.
    """
    if isinstance(
        fp,
        (
            UIntSparseIntVect,
            IntSparseIntVect,
            LongSparseIntVect,
            ULongSparseIntVect,
        ),
    ):
        tmp = fp.GetNonzeroElements()

    elif isinstance(fp, SparseBitVect):
        on_bits = fp.GetOnBits()
        tmp = dict(zip(on_bits, np.ones(len(on_bits))))

    else:
        raise ValueError(f"The fingerprint is of wrong type: {type(fp)}")

    # ON bits dict to (i, v)
    i = np.array(list(tmp.keys())) % dim
    v = np.array(list(tmp.values()))

    # Fold indices
    i = i % dim

    # Create the folded fp
    folded = np.zeros(dim, dtype="int")
    np.add.at(folded, i, v)

    if binary:
        folded = np.clip(folded, a_min=0, a_max=1)

    return folded



import pytest

import datamol as dm


def check_logs_are_shown(capfd):
    smiles = "fake_smiles"
    dm.to_mol(smiles)
    _, err = capfd.readouterr()
    assert "SMILES Parse Error" in err


def check_logs_are_not_shown(capfd):
    smiles = "fake_smiles"
    dm.to_mol(smiles)
    _, err = capfd.readouterr()
    assert err == ""


@pytest.mark.skip_platform("win")
def test_rdkit_log(capfd):
    """Test multiple rdkit log scenarios."""

    check_logs_are_shown(capfd)
    with dm.without_rdkit_log():
        check_logs_are_not_shown(capfd)
    check_logs_are_shown(capfd)

    dm.disable_rdkit_log()
    check_logs_are_not_shown(capfd)

    dm.enable_rdkit_log()
    check_logs_are_shown(capfd)

    dm.disable_rdkit_log()
    with dm.without_rdkit_log():
        check_logs_are_not_shown(capfd)
    check_logs_are_not_shown(capfd)


@pytest.mark.skip_platform("win")
def test_rdkit_log_enable(capfd):
    dm.enable_rdkit_log()

    with dm.without_rdkit_log():
        check_logs_are_not_shown(capfd)

    with dm.without_rdkit_log(enable=False):
        check_logs_are_shown(capfd)

    check_logs_are_shown(capfd)



from typing import Any
from typing import Union
from typing import List

import datamol as dm

from ._viz import to_image


def match_substructure(
    mols: Union[List[dm.Mol], dm.Mol],
    queries: Union[List[dm.Mol], dm.Mol],
    highlight_bonds: bool = True,
    copy: bool = True,
    **kwargs: Any,
):
    """Generate an image of molecule(s) with substructure matches for a given
    pattern or substructure.

    Args:
        mols: One or more molecules.
        queries: One or more queries.
        highlight_bonds: Whether to also highlight the bonds matching the patterns.
        copy: Whether to copy the molecules and the queries.
        kwargs: Other kwargs passed to `dm.viz.to_image`.
    """

    # NOTE(hadim): `MolsToGridImage` used in `to_image` can't use a list of list of indices
    # for every molecules so it's not really possible to have different colors for different
    # matches in the same molecules.
    # In the future, we will implement our custom `MolsToGridImage` in order to have more controls
    # on the colors used.
    # For the same reason, we don't bother about colors here.

    if isinstance(mols, dm.Mol):
        mols = [mols]

    if isinstance(queries, dm.Mol):
        queries = [queries]

    # Copy mols and patterns
    if copy:
        mols = [dm.copy_mol(mol) for mol in mols]
        queries = [dm.copy_mol(mol) for mol in queries]

    all_atom_indices = []
    all_bond_indices = []

    for mol in mols:
        atom_indices = []
        bond_indices = []

        for query in queries:
            if highlight_bonds:
                atom_matches, bond_matches = dm.substructure_matching_bonds(mol, query)
                atom_indices += atom_matches
                bond_indices += bond_matches
            else:
                atom_indices += list(mol.GetSubstructMatches(query, uniquify=True))  # type: ignore
                bond_indices += []

        # NOTE(hadim): we must flatten the atom/bond indices, since `MolsToGridImage`
        # don't accept multiple list of indices for every single molecule.
        bond_indices = [item for sublist in bond_indices for item in sublist]
        atom_indices = [item for sublist in atom_indices for item in sublist]

        all_atom_indices.append(atom_indices)
        all_bond_indices.append(bond_indices)

    image = to_image(
        mols,
        highlight_atom=all_atom_indices,
        highlight_bond=all_bond_indices,
        **kwargs,
    )

    return image



from typing import Set
from typing import Optional
from typing import Any

from rdkit import Chem
from rdkit.Chem import BRICS
from rdkit.Chem import Recap
from rdkit.Chem import rdMMPA

from rdkit.Chem.Fraggle import FraggleSim

import datamol as dm


def brics(
    mol: Chem.rdchem.Mol,
    singlepass: bool = True,
    remove_parent: bool = False,
    sanitize: bool = True,
    fix: bool = True,
):
    """Run BRICS on the molecules and potentially fix dummy atoms.

    Args:
        mol: a molecule.
        singlepass: Single pass for `BRICSDecompose`.
        remove_parent: Remove parent from the fragments.
        sanitize: Wether to sanitize the fragments.
        fix: Wether to fix the fragments.
    """
    frags = BRICS.BRICSDecompose(mol, returnMols=True, singlePass=singlepass)
    frags = list(frags)

    if fix:
        frags = [dm.fix_mol(x) for x in frags]
    if sanitize:
        frags = [dm.sanitize_mol(x) for x in frags]
    if remove_parent:
        frags.pop(0)

    frags = [x for x in frags if x is not None]

    return frags


def frag(
    mol: Chem.rdchem.Mol,
    remove_parent: bool = False,
    sanitize: bool = True,
    fix: bool = True,
):
    """Generate all possible fragmentation of a molecule.

    Args:
        mol: a molecule.
        remove_parent: Remove parent from the fragments.
        sanitize: Wether to sanitize the fragments.
        fix: Wether to fix the fragments.
    """
    frags = FraggleSim.generate_fraggle_fragmentation(mol)

    smiles = set([])
    for seq in frags:
        smiles |= {s.strip() for s in seq.split(".")}

    smiles = list(sorted(smiles, reverse=True))
    frags = [dm.to_mol(s) for s in smiles]

    if fix:
        frags = [dm.fix_mol(x) for x in frags]
    if sanitize:
        frags = [dm.sanitize_mol(x) for x in frags]

    frags = [x for x in frags if x is not None]

    if remove_parent:
        return frags
    return [mol] + frags


def recap(
    mol: Chem.rdchem.Mol,
    remove_parent: bool = False,
    sanitize: bool = True,
    fix: bool = True,
):
    """Fragment the molecule using the recap algorithm.

    Args:
        mol: a molecule.
        remove_parent: Remove parent from the fragments.
        sanitize: Wether to sanitize the fragments.
        fix: Wether to fix the fragments.
    """
    res = Recap.RecapDecompose(mol)
    frags = [dm.to_mol(x) for x in res.GetAllChildren().keys()]

    if fix:
        frags = [dm.fix_mol(x) for x in frags]
    if sanitize:
        frags = [dm.sanitize_mol(x) for x in frags]

    frags = [x for x in frags if x is not None]

    if remove_parent:
        return frags
    return [mol] + frags


def anybreak(
    mol: Chem.rdchem.Mol,
    remove_parent: bool = False,
    sanitize: bool = True,
    fix: bool = True,
):
    """Fragment molecule by applying brics first, then fall back to frag.

    Args:
        mol: a molecule.
        remove_parent: Remove parent from the fragments.
        sanitize: Wether to sanitize the fragments.
        fix: Wether to fix the fragments.
    """
    frags = []
    try:
        frags = brics(mol, fix=fix, remove_parent=remove_parent, sanitize=sanitize)
    except:
        pass

    if len(frags) == 0:
        frags = frag(mol, remove_parent=remove_parent, sanitize=sanitize, fix=fix)

    return frags


def mmpa_frag(
    mol: dm.Mol,
    pattern: Optional[str] = None,
    max_cut: int = 1,
    max_bond_cut: int = 20,
    h_split: bool = False,
) -> Optional[Set[dm.Mol]]:
    """Fragment molecule on specific bonds suitable for a MMPA analysis.

    Args:
        mol: Molecule to fragment.
        pattern: Bond pattern to split on. Will use default rdkit pattern
            '[#6+0;!$(*=,#[!#6])]!@!=!#[*]' if not provided.
        max_cut: Number of cuts.
        max_bond_cut: Maximum number of bond to cut. Default to 20.
        h_split:  Whether to split at hydrogen position too.
            This is equivalent to enabling the addition of new fragments.

    Returns:
        List of fragments.
    """

    frags = []
    if pattern is None:
        frags = rdMMPA.FragmentMol(
            mol,
            maxCuts=max_cut,
            resultsAsMols=False,
            maxCutBonds=max_bond_cut,
        )
    elif pattern:
        frags = rdMMPA.FragmentMol(
            mol,
            pattern=pattern,
            maxCuts=max_cut,
            resultsAsMols=False,
            maxCutBonds=max_bond_cut,
        )

    if h_split:
        mol = dm.add_hs(mol)
        frags += rdMMPA.FragmentMol(
            mol,
            pattern="[#1]!@!=!#[!#1]",
            maxCuts=1,
            resultsAsMols=False,
            maxCutBonds=max_bond_cut,
        )
    return set(frags)


def mmpa_cut(mol: dm.Mol, rdkit_pattern: bool = False) -> Optional[Set[Any]]:
    """Cut molecules to perform mmpa analysis later

    Args:
        mol: Molecule to fragment.
        rdkit_pattern: Whether to perform the fragmentation
            using the default rdkit pattern: [#6+0;!$(*=, #[!#6])]!@!=!#[*]"

    Returns:
        List of 'smiles,core,chains'
    """

    if mol is None:
        return mol

    outlines = set()

    smiles = dm.to_smiles(mol)

    if rdkit_pattern:
        frags = mmpa_frag(mol, max_cut=3, max_bond_cut=30)
    else:
        # heavy atoms
        frags = mmpa_frag(mol, pattern="[!#1]!@!=!#[!#1]", max_cut=4, max_bond_cut=30)
        frags.update(mmpa_frag(mol, pattern="[!#1]!@!=!#[!#1]", max_cut=3, max_bond_cut=30))

    frags = set(frags)
    for core, chains in frags:
        output = f"{smiles},{core},{chains}\n"
        outlines.add(output)

    # hydrogen splitting
    mol = dm.add_hs(mol)
    smiles = dm.to_smiles(mol)

    n = mol.GetNumHeavyAtoms()
    if n < 60:
        frags = mmpa_frag(mol, pattern=None, max_cut=1, max_bond_cut=100, h_split=True)
        for core, chains in frags:
            output = f"{smiles},{core},{chains}\n"
            outlines.add(output)

    return outlines



from ._fragment import brics
from ._fragment import frag
from ._fragment import recap
from ._fragment import anybreak
from ._fragment import mmpa_frag
from ._fragment import mmpa_cut

from ._assemble import assemble_fragment_iter
from ._assemble import assemble_fragment_order
from ._assemble import break_mol
from ._assemble import build



"""The `fs` module makes it easier to work with all type of path (the ones supported by `fsspec`).
"""

from typing import Union
from typing import Optional
from typing import List

import os
import io
import hashlib
import pathlib

import platformdirs

import fsspec
import fsspec.utils

from .jobs import parallelized


def _import_tqdm():
    try:
        from tqdm.auto import tqdm

        return tqdm
    except ImportError:
        return None


def get_cache_dir(app_name: str, suffix: Optional[str] = None, create: bool = True):
    """Get a local cache directory for a given application name.

    Args:
        app_name: The name of the application.
        suffix: A subdirectory appended to the cache dir.
        create: Whether to create the directory and its parents if it does not
            already exist.
    """

    cache_dir = pathlib.Path(platformdirs.user_cache_dir(appname=app_name))

    if suffix is not None:
        cache_dir /= suffix

    if create:
        cache_dir.mkdir(exist_ok=True, parents=True)

    return cache_dir


def get_mapper(path: Union[str, os.PathLike]):
    """Get the fsspec mapper.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    return fsspec.get_mapper(str(path))


def get_basename(path: Union[str, os.PathLike]):
    """Get the basename of a file or a folder.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    path = str(path)
    mapper = get_mapper(path)
    clean_path = path.rstrip(mapper.fs.sep)
    return str(clean_path).split(mapper.fs.sep)[-1]


def get_extension(path: Union[str, os.PathLike]):
    """Get the extension of a file.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    basename = get_basename(path)
    return basename.split(".")[-1]


def exists(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file or a directory exists.

    Important: File-like object always exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    return is_file(path) or is_dir(path)


def is_file(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    if isinstance(path, fsspec.core.OpenFile):
        return path.fs.isfile(path.path)

    elif isinstance(path, (str, os.PathLike)):
        mapper = get_mapper(str(path))
        return mapper.fs.isfile(str(path))

    else:
        return False


def is_dir(path: Union[str, os.PathLike, fsspec.core.OpenFile, io.IOBase]):
    """Check whether a file exists.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """
    if isinstance(path, fsspec.core.OpenFile):
        return path.fs.isdir(path.path)

    elif isinstance(path, (str, os.PathLike)):
        mapper = get_mapper(str(path))
        return mapper.fs.isdir(str(path))

    else:
        return False


def get_protocol(path: Union[str, os.PathLike], fs: Optional[fsspec.AbstractFileSystem] = None):
    """Return the name of the path protocol.

    Args:
        path: a path supported by `fsspec` such as local, s3, gcs, etc.
    """

    if fs is None:
        fs = get_mapper(path).fs

    protocol = fs.protocol  # type: ignore

    if "s3" in protocol:
        return "s3"
    elif "gs" in protocol:
        return "gs"
    elif isinstance(protocol, (tuple, list)):
        return protocol[0]
    return protocol


def is_local_path(path: Union[str, os.PathLike]):
    """Check whether a path is local."""
    return get_protocol(str(path)) == "file"


def join(*paths: str):
    """Join paths together. The first element determine the
    filesystem to use (and so the separator.

    Args:
        *paths: a list of paths supported by `fsspec` such as local, s3, gcs, etc.
    """
    _paths = [str(path).rstrip("/") for path in paths]
    source_path = _paths[0]
    fs = get_mapper(source_path).fs
    full_path = fs.sep.join(_paths)
    return full_path


def get_size(file: Union[str, os.PathLike, io.IOBase, fsspec.core.OpenFile]) -> Optional[int]:
    """Get the size of a file given its path. Return None if the
    size can't be retrieved.
    """

    if isinstance(file, io.IOBase) and hasattr(file, "name"):
        fs_local = fsspec.filesystem("file")
        file_size = fs_local.size(getattr(file, "name"))

    elif isinstance(file, (str, os.PathLike)):
        fs = get_mapper(str(file)).fs
        file_size = fs.size(str(file))

    elif isinstance(file, fsspec.core.OpenFile):
        file_size = file.fs.size(file.path)

    else:
        file_size = None

    return file_size


def copy_file(
    source: Union[str, pathlib.Path, io.IOBase, fsspec.core.OpenFile],
    destination: Union[str, pathlib.Path, io.IOBase, fsspec.core.OpenFile],
    chunk_size: Optional[int] = None,
    force: bool = False,
    progress: bool = False,
    leave_progress: bool = True,
):
    """Copy one file to another location across different filesystem (local, S3, GCS, etc).

    Args:
        source: path or file-like object to copy from.
        destination: path or file-like object to copy to.
        chunk_size: the chunk size to use. If progress is enabled the chunk
            size is `None`, it is set to 1MB (1024 * 1024).
        force: whether to overwrite the destination file if it exists.
        progress: whether to display a progress bar.
        leave_progress: whether to hide the progress bar once the copy is done.
    """

    if progress and chunk_size is None:
        chunk_size = 1024 * 1024

    if isinstance(source, (str, os.PathLike)):
        source_file = fsspec.open(str(source), "rb")
    else:
        source_file = source

    if isinstance(destination, (str, os.PathLike)):
        # adapt the file mode of the destination depending on the source file.
        destination_mode = "wb"
        if hasattr(source_file, "mode"):
            destination_mode = "wb" if "b" in getattr(source_file, "mode") else "w"
        elif isinstance(source_file, io.BytesIO):
            destination_mode = "wb"
        elif isinstance(source_file, io.StringIO):
            destination_mode = "w"

        destination_file = fsspec.open(str(destination), destination_mode)
    else:
        destination_file = destination

    if not is_file(source_file):  # type: ignore
        raise ValueError(f"The file being copied does not exist or is not a file: {source}")

    if not force and is_file(destination_file):  # type: ignore
        raise ValueError(f"The destination file to copy already exists: {destination}")

    with source_file as source_stream:
        with destination_file as destination_stream:
            if chunk_size is None:
                # copy without chunks
                destination_stream.write(source_stream.read())  # type: ignore

            else:
                # copy with chunks

                # determine the size of the source file
                source_size = None
                if progress:
                    source_size = get_size(source)

                pbar = None
                if progress:
                    tqdm = _import_tqdm()

                    if tqdm is None:
                        raise ImportError(
                            "If the progress bar is enabled, you must have `tqdm` "
                            "installed: `conda install tqdm`."
                        )
                    else:
                        # init progress bar
                        pbar = tqdm(
                            total=source_size,
                            leave=leave_progress,
                            disable=not progress,
                            unit="B",
                            unit_divisor=1024,
                            unit_scale=True,
                        )

                # start the loop
                while True:
                    data = source_stream.read(chunk_size)  # type: ignore
                    if not data:
                        break
                    destination_stream.write(data)  # type: ignore

                    if pbar is not None:
                        pbar.update(chunk_size)

                if pbar is not None:
                    pbar.close()


def mkdir(dir_path: Union[str, os.PathLike], exist_ok: bool = False):
    """Create a directory.

    Args:
        dir_path: The path of the directory to create.
        exist_ok: Whether to ignore the error if the directory
            already exists.
    """
    fs = get_mapper(str(dir_path)).fs
    fs.mkdirs(str(dir_path), exist_ok=exist_ok)


def md5(filepath: Union[str, os.PathLike]):
    """Return the md5 hash of a file.

    Args:
        filepath: The path to the file to compute the MD5 hash on.
    """
    with fsspec.open(filepath) as f:
        file_hash = hashlib.md5()
        file_hash.update(f.read())  # type: ignore
        file_hash = file_hash.hexdigest()
    return file_hash


def glob(path: str, detail: bool = False, **kwargs) -> List[str]:
    """Find files by glob-matching.

    Args:
        path: A glob-style path.
    """
    # Get the list of paths
    fs = get_mapper(path).fs
    paths = fs.glob(path, detail=detail, **kwargs)
    paths = [fsspec.utils._unstrip_protocol(d, fs) for d in paths]
    return paths


def copy_dir(
    source: Union[str, pathlib.Path],
    destination: Union[str, pathlib.Path],
    force: bool = False,
    progress: bool = False,
    leave_progress: bool = True,
    file_progress: bool = False,
    file_leave_progress: bool = False,
    chunk_size: Optional[int] = None,
):
    """Copy one directory to another location across different filesystem (local, S3, GCS, etc).

    Note that if both FS from source and destination are the same, progress won't be shown.

    Args:
        source: Path to the source directory.
        destination: Path to the destination directory.
        chunk_size: the chunk size to use. If progress is enabled the chunk
            size is `None`, it is set to 2048.
        force: whether to overwrite the destination directory if it exists.
        progress: Whether to display a progress bar.
        leave_progress: Whether to hide the progress bar once the copy is done.
        file_progress: Whether to display a progress bar for each file.
        file_leave_progress: Whether to hide the progress bar once a file copy is done.
        chunk_size: See `dm.utils.fs.copy_file`.
    """

    source = str(source)
    destination = str(destination)

    source_fs = get_mapper(source).fs
    destination_fs = get_mapper(destination).fs

    # Sanity check
    if not is_dir(source):
        raise ValueError(
            f"The directory being copied does not exist or is not a directory: {source}"
        )

    if not force and is_dir(destination):
        raise ValueError(f"The destination folder to copy already exists: {destination}")

    # If both fs are the same then we just rely on the internal `copy` method
    # which is much faster.
    if destination_fs.__class__ == source_fs.__class__:
        source_fs.copy(source, destination, recursive=True)
        return

    # Get all input paths with details
    # NOTE(hadim): we could have use `.glob(..., detail=True)` here but that API is inconsistent
    # between the backends resulting in different object types being returned (dict, list, etc).
    detailed_paths = source_fs.find(source, withdirs=True, detail=True)
    detailed_paths = list(detailed_paths.values())

    # Get list of input types
    input_types = [d["type"] for d in detailed_paths]

    # Get list of input path + add protocol if needed
    input_paths = [d["name"] for d in detailed_paths]
    input_paths = [fsspec.utils._unstrip_protocol(p, source_fs) for p in input_paths]

    # Build all the output paths
    output_paths: List[str] = fsspec.utils.other_paths(input_paths, destination)  # type: ignore

    def _copy_source_to_destination(input_path, input_type, output_path):
        # A directory
        if input_type == "directory":
            destination_fs.mkdir(output_path)

        # A file
        else:
            copy_file(
                input_path,
                output_path,
                force=force,
                progress=file_progress,
                leave_progress=file_leave_progress,
                chunk_size=chunk_size,
            )

    # Copy source files/directories to destination in parallel
    parallelized(
        _copy_source_to_destination,
        inputs_list=list(zip(input_paths, input_types, output_paths)),
        arg_type="args",
        progress=progress,
        tqdm_kwargs=dict(leave=leave_progress),
        scheduler="threads",
    )



from ._reactions import is_reaction_ok
from ._reactions import select_reaction_output
from ._reactions import apply_reaction
from ._reactions import can_react
from ._reactions import inverse_reaction
from ._reactions import find_reactant_position
from ._reactions import ATTACHING_RXN
from ._reactions import rxn_from_smarts
from ._reactions import rxn_to_smarts
from ._reactions import rxn_from_block
from ._reactions import rxn_from_block_file
from ._reactions import rxn_to_block
from ._reactions import rxn_to_block_file

from ._attachments import add_brackets_to_attachment_points
from ._attachments import convert_attach_to_isotope
from ._attachments import num_attachment_points
from ._attachments import open_attach_points



from typing import Union
from typing import List

import copy
import itertools

from rdkit import Chem
from rdkit.Chem import rdMolAlign


def _get_nglview():
    try:
        import nglview as nv

        return nv
    except ImportError:
        raise ImportError("You must install nglview from https://github.com/nglviewer/nglview.")


def _get_ipywidgets():
    try:
        import ipywidgets as widgets

        return widgets
    except ImportError:
        raise ImportError(
            "You must install ipywidgets from https://github.com/jupyter-widgets/ipywidgets/."
        )


def conformers(
    mol: Chem.rdchem.Mol,
    conf_id: int = -1,
    n_confs: Union[int, List[int]] = None,
    align_conf: bool = True,
    n_cols: int = 3,
    sync_views: bool = True,
    remove_hs: bool = True,
    width: str = "auto",
):
    """Visualize the conformer(s) of a molecule.

    Args:
        mol: a molecule.
        conf_id: The ID of the conformer to show. -1 shows
            the first conformer. Only works if `n_confs` is None.
        n_confs: Can be a number of conformers
            to shows or a list of conformer indices. When None, only the first
            conformer is displayed. When -1, show all conformers.
        align_conf: Whether to align conformers together.
        n_cols: Number of columns. Defaults to 3.
        sync_views: Wether to sync the multiple views.
        remove_hs: Wether to remove the hydrogens of the conformers.
        width: The width of the returned view. Defaults to "auto".
    """

    widgets = _get_ipywidgets()
    nv = _get_nglview()

    if mol.GetNumConformers() == 0:
        raise ValueError(
            "The molecule has 0 conformers. You can generate conformers with `dm.conformers.generate(mol)`."
        )

    # Clone the molecule
    mol = copy.deepcopy(mol)

    if remove_hs:
        mol = Chem.RemoveHs(mol)  # type: ignore
    else:
        mol = Chem.AddHs(mol)  # type: ignore

    if n_confs is None:
        return nv.show_rdkit(mol, conf_id=conf_id)

    # If n_confs is int, convert to list of conformer IDs
    if n_confs == -1:
        n_confs = [conf.GetId() for conf in mol.GetConformers()]
    elif isinstance(n_confs, int):
        if n_confs > mol.GetNumConformers():
            n_confs = mol.GetNumConformers()
        n_confs = list(range(n_confs))  # type: ignore

    if align_conf:
        rdMolAlign.AlignMolConformers(mol, confIds=n_confs)

    # Get number of rows
    n_rows = len(n_confs) // n_cols
    n_rows += 1 if (len(n_confs) % n_cols) > 0 else 0

    # Create a grid
    grid = widgets.GridspecLayout(n_rows, n_cols)  # type: ignore

    # Create and add views to the grid.
    widget_coords = itertools.product(range(n_rows), range(n_cols))
    views = []
    for i, (conf_id, (x, y)) in enumerate(zip(n_confs, widget_coords)):
        view = nv.show_rdkit(mol, conf_id=conf_id)
        view.layout.width = width
        view.layout.align_self = "stretch"
        grid[x, y] = view
        views.append(view)

    # Sync views
    if sync_views:
        for view in views:
            view._set_sync_camera(views)

    return grid



from rdkit import RDLogger
from rdkit import rdBase


class without_rdkit_log:
    """Context manager to disable RDKit logs. By default all logs are disabled.

    Example:

    ```python
    import datamol as dm

    with dm.without_rdkit_log():
        mol = dm.to_mol("CCCCO")  # potential RDKit logs won't show
    ```
    """

    def __init__(
        self,
        mute_errors: bool = True,
        mute_warning: bool = True,
        mute_info: bool = True,
        mute_debug: bool = True,
        enable: bool = True,
    ):
        if enable is False:
            mute_errors = False
            mute_warning = False
            mute_info = False
            mute_debug = False

        # Get current log state
        self.previous_status = self._get_log_status()

        # Init the desired log state to apply during in the context
        self.desired_status = {}
        self.desired_status["rdApp.error"] = not mute_errors
        self.desired_status["rdApp.warning"] = not mute_warning
        self.desired_status["rdApp.debug"] = not mute_debug
        self.desired_status["rdApp.info"] = not mute_info

    def _get_log_status(self):
        """Get the current log status of RDKit logs."""
        log_status = rdBase.LogStatus()
        log_status = {st.split(":")[0]: st.split(":")[1] for st in log_status.split("\n")}
        log_status = {k: True if v == "enabled" else False for k, v in log_status.items()}
        return log_status

    def _apply_log_status(self, log_status):
        """Apply an RDKit log status."""
        for k, v in log_status.items():
            if v is True:
                rdBase.EnableLog(k)
            else:
                rdBase.DisableLog(k)

    def __enter__(self):
        self._apply_log_status(self.desired_status)

    def __exit__(self, *args, **kwargs):
        self._apply_log_status(self.previous_status)


def disable_rdkit_log():
    """Disable all rdkit logs."""
    for log_level in RDLogger._levels:
        rdBase.DisableLog(log_level)


def enable_rdkit_log():
    """Enable all rdkit logs."""
    for log_level in RDLogger._levels:
        rdBase.EnableLog(log_level)



import datamol as dm


def test_fuzzy_scaffolding():
    smiles = [
        "Cc1ccc(NC(=O)Cn2cccn2)c(Br)c1",
        "COc1ccc(OC(C)C(=O)N=c2sccn2C)cc1",
        "CC(NC(=O)CSCc1cccs1)C1CCCO1",
        "CC1CCCCN1C(=O)CN1CCC[C@@H](N)C1",
        "CCC(CC)COC(=O)[C@H](C)N[P@](=O)(OC[C@H]1O[C@](C#N)([C@H](O)[C@@H]1O)C1=CC=C2N1N=CN=C2N)OC1=CC=CC=C1",  # no way this one (Remdesivir) is in the db
    ]

    mols = [dm.to_mol(s) for s in smiles]
    all_scaffolds, scf2infos, scf2groups = dm.scaffold.fuzzy_scaffolding(mols)

    assert scf2infos.keys() == scf2groups.keys()
    assert len(all_scaffolds) == 5

    # NOTE(hadim): different version of rdkit (2020.09 vs 2021.03) returns
    # different SMILES here.
    # assert "O=C(CN1CCC[C@@H]([*:1])C1)N1CCCCC1[*:2]" in all_scaffolds
    # assert "O=C(CSCc1cccs1)NC(C1CCCO1)[*:1]" in all_scaffolds
    # assert "O=C(N=c1sccn1[*:1])C(Oc1ccc([*:3])cc1)[*:2]" in all_scaffolds



import datamol as dm
import unittest as ut


def test_to_mol():
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    graph = dm.to_graph(mol)

    assert list(graph.nodes) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    assert list(graph.edges) == [
        (0, 1),
        (1, 2),
        (1, 3),
        (3, 4),
        (4, 5),
        (4, 9),
        (5, 6),
        (6, 7),
        (7, 8),
        (8, 9),
        (9, 10),
        (10, 11),
        (10, 12),
    ]


def test_get_all_path_between():
    smiles = "c1cc2cccccc2c1"
    mol = dm.to_mol(smiles)

    all_paths = dm.get_all_path_between(mol, 8, 4, ignore_cycle_basis=False)
    assert all_paths == [[8, 2, 3, 4], [8, 7, 6, 5, 4], [8, 9, 0, 1, 2, 3, 4]]

    all_paths = dm.get_all_path_between(mol, 8, 4, ignore_cycle_basis=True)
    assert all_paths == [[8, 2, 3, 4], [8, 7, 6, 5, 4]]


class Test_match_molecular_graphs(ut.TestCase):
    def test_ring(self):
        mol1 = dm.to_mol("C1CCCCC1", ordered=False)
        mol2 = dm.to_mol("C1=CC=CC=C1", ordered=True)
        mol3 = dm.to_mol("C1CCOCC1", ordered=False)
        mol4 = dm.to_mol("C1=CCOCC1", ordered=True)
        mol5 = dm.to_mol("C1=CCCCC1", ordered=True)
        mol6 = dm.to_mol("C1CCCCC1", ordered=True, add_hs=True)

        # 12 matches on rings (6 rotations * 2 reflections)
        matches = dm.match_molecular_graphs(
            mol1, mol1, match_atoms_on="atomic_num", match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 12)

        # 12 matches on rings (6 rotations * 2 reflections)
        matches = dm.match_molecular_graphs(
            mol1, mol2, match_atoms_on="atomic_num", match_bonds_on=[]
        )
        self.assertEqual(len(matches), 12)

        # 12 matches on rings (6 rotations * 2 reflections)
        matches = dm.match_molecular_graphs(
            mol2, mol2, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 12)

        # 0 matches on rings of different bond types
        matches = dm.match_molecular_graphs(
            mol1, mol2, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 0)

        # Matches with Oxygen atom
        matches = dm.match_molecular_graphs(
            mol1, mol3, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 12)

        # No matching due to atom type
        matches = dm.match_molecular_graphs(
            mol1, mol3, match_atoms_on=["atomic_num"], match_bonds_on=[]
        )
        self.assertEqual(len(matches), 0)

        # 2 reflections can match
        matches = dm.match_molecular_graphs(
            mol3, mol4, match_atoms_on=["atomic_num"], match_bonds_on=[]
        )
        self.assertEqual(len(matches), 2)

        # No matching due to different bonds
        matches = dm.match_molecular_graphs(
            mol3, mol4, match_atoms_on=["atomic_num"], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 0)

        # 2 reflections can match
        matches = dm.match_molecular_graphs(
            mol4, mol5, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 2)

        # 0 matches with hydrogens
        matches = dm.match_molecular_graphs(
            mol1, mol6, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 0)

        # 768 matches with hydrogens when specifying explicit_hs.
        # 768 = 6 rotations * 2 reflections * (2 reflection per carbon with Hs = 2^6)

        matches = dm.match_molecular_graphs(
            dm.add_hs(mol1), mol6, match_atoms_on=[], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 768)

    def test_mol(self):
        smiles1 = (
            "O=C(CC[C@@H]1NC(=O)N(CCc2c[nH]c3ccccc23)C1=O)N1C[C@@H]2C[C@H](C1)[C]1C=C[CH]C(=O)N1C2"
        )
        mol1 = dm.to_mol(smiles1, ordered=False)
        mol1_ordered = dm.to_mol(smiles1, ordered=True)

        smiles2 = "O=C1C2=C(N=CN2C)N(C(=O)N1C)C"
        mol2 = dm.to_mol(smiles2, ordered=False)
        mol2_ordered = dm.to_mol(smiles2, ordered=True)

        # Single match. Check if re-ordered atomic number match
        matches = dm.match_molecular_graphs(
            mol1, mol1_ordered, match_atoms_on=["atomic_num"], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 1)
        match = matches[0]
        atoms1 = [atom.GetAtomicNum() for atom in mol1.GetAtoms()]
        atoms1_ordered = [atom.GetAtomicNum() for atom in mol1_ordered.GetAtoms()]
        atoms1_re_ordered = [atoms1_ordered[match[ii]] for ii in range(len(match))]
        self.assertListEqual(atoms1, atoms1_re_ordered)

        # Single match. Check if re-ordered atomic number match
        matches = dm.match_molecular_graphs(
            mol2, mol2_ordered, match_atoms_on=["atomic_num"], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 1)
        match = matches[0]
        atoms2 = [atom.GetAtomicNum() for atom in mol2.GetAtoms()]
        atoms2_ordered = [atom.GetAtomicNum() for atom in mol2_ordered.GetAtoms()]
        atoms2_re_ordered = [atoms2_ordered[match[ii]] for ii in range(len(match))]
        self.assertListEqual(atoms2, atoms2_re_ordered)

        # Molecules don't match
        matches = dm.match_molecular_graphs(
            mol1, mol2, match_atoms_on=["atomic_num"], match_bonds_on=["bond_type"]
        )
        self.assertEqual(len(matches), 0)


class Test_reorder_mol_from_template(ut.TestCase):
    def test_reorder_mol_from_template(self):
        # No re-ordering because too many matches
        mol1 = dm.to_mol("C1CCCCC1", ordered=False)
        mol2 = dm.to_mol("C1=CC=CC=C1", ordered=True)
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2)
        self.assertIsNone(mol_reordered)

        # Re-ordering with option ambiguous_match_mode
        mol1 = dm.to_mol("C1CCCCC1", ordered=False)
        mol2 = dm.to_mol("C1=CC=CC=C1", ordered=True)
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2, ambiguous_match_mode="first")
        self.assertIsInstance(mol_reordered, dm.Mol)
        self.assertTrue(dm.same_mol(mol_reordered, mol1))

        # Re-ordering with option allow_ambiguous_hs_only should work since only hydrogens are ambiguous
        mol1 = dm.add_hs(dm.to_mol("C[N]1C=NC2=C1C(=O)N(C(=O)N2C)C", ordered=False))
        mol2 = dm.add_hs(dm.to_mol("C[N]1C=NC2=C1C(=O)N(C(=O)N2C)C", ordered=True))
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2, ambiguous_match_mode="hs-only")
        self.assertIsInstance(mol_reordered, dm.Mol)
        self.assertTrue(dm.same_mol(mol_reordered, mol1))

        # Re-ordering with option allow_ambiguous_hs_only should not work since number of hydrogens are ambiguous
        mol1 = dm.add_hs(dm.to_mol("C1CCCCC1", ordered=False))
        mol2 = dm.add_hs(dm.to_mol("C1CCCCC1", ordered=True))
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2, ambiguous_match_mode="hs-only")
        self.assertIsNone(mol_reordered)

        # Re-ordering with option allow_ambiguous_hs_only should not work since some non-hydrogens are ambiguous
        mol1 = dm.add_hs(dm.to_mol("C1CCCCC1", ordered=False))
        mol2 = dm.add_hs(dm.to_mol("C1=CC=CC=C1", ordered=True))
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2, ambiguous_match_mode="hs-only")
        self.assertIsNone(mol_reordered)

        # Re-ordering with option ambiguous_match_mode=="best" should work
        mol1 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)CC)C(=O)O", ordered=False)
        mol2 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)C=C)C(=O)O", ordered=True)
        mol_reordered = dm.reorder_mol_from_template(
            mol1,
            mol2,
            enforce_atomic_num=True,
            enforce_bond_type=False,
            ambiguous_match_mode="best",
        )
        self.assertIsInstance(mol_reordered, dm.Mol)

        # Re-ordering with option ambiguous_match_mode=="best" should not work due to hydrogens
        mol1 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)CC)C(=O)O", ordered=False)
        mol2 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)C=C)C(=O)O", ordered=True)
        mol1, mol2 = dm.add_hs(mol1), dm.add_hs(mol2)
        mol_reordered = dm.reorder_mol_from_template(
            mol1,
            mol2,
            enforce_atomic_num=True,
            enforce_bond_type=False,
            ambiguous_match_mode="best",
        )
        self.assertIsNone(mol_reordered)

        # Re-ordering works with best, followed by first should work
        mol1 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)CC)C(=O)O", ordered=False)
        mol2 = dm.to_mol("O=C(C)OC1=C(C=C(C=C1)C=C)C(=O)O", ordered=True)
        mol_reordered = dm.reorder_mol_from_template(
            mol1,
            mol2,
            enforce_atomic_num=True,
            enforce_bond_type=False,
            ambiguous_match_mode="best-first",
        )
        self.assertIsInstance(mol_reordered, dm.Mol)
        self.assertEqual(
            sum(
                [
                    bond1.GetBondType() != bond2.GetBondType()
                    for bond1, bond2 in zip(mol_reordered.GetBonds(), mol2.GetBonds())
                ]
            ),
            1,
        )
        self.assertEqual(
            sum(
                [
                    atom1.GetAtomicNum() != atom2.GetAtomicNum()
                    for atom1, atom2 in zip(mol_reordered.GetAtoms(), mol2.GetAtoms())
                ]
            ),
            0,
        )

        # No-reordering because no matches
        mol1 = dm.to_mol("C1CCCCC1", ordered=False)
        mol2 = dm.to_mol("C1=CC=CC=C1", ordered=True)
        mol_reordered = dm.reorder_mol_from_template(mol1, mol2, enforce_bond_type=True)
        self.assertIsNone(mol_reordered)

        smiles3 = "O=C(CC[C@@H]3NC(=O)N(CCC1=C[N]([H])C2=C1C=CC=C2)C3=O)N4C[C@@H]6C[C@H](C4)C5C=CC([H])C(=O)N5C6"
        smiles3_variation = "O=C(CC[C@@H]3NC(=O)N(CCC1=C[N]([H])C2=C1C[N]CC2)C3=O)N4C[C@@H]6C[C@H](C4)C5C=CC([H])C(=O)N5C6"
        mol3 = dm.to_mol(smiles3, ordered=False)
        mol3_ordered = dm.to_mol(smiles3, ordered=True)
        mol3_variation = dm.to_mol(smiles3_variation, ordered=True)

        # Check re-ordering of molecules without enforcing
        mol3_reordered = dm.reorder_mol_from_template(
            mol3, mol3_ordered, enforce_atomic_num=True, enforce_bond_type=False
        )
        self.assertIsInstance(mol3_reordered, dm.Mol)
        atoms3 = [atom.GetAtomicNum() for atom in mol3.GetAtoms()]
        atoms3_ordered = [atom.GetAtomicNum() for atom in mol3_ordered.GetAtoms()]
        atoms3_reordered = [atom.GetAtomicNum() for atom in mol3_reordered.GetAtoms()]
        self.assertListEqual(atoms3_ordered, atoms3_reordered)
        self.assertFalse(all([atoms3[ii] == atoms3_ordered[ii] for ii in range(len(atoms3))]))

        # Check re-ordering of molecules without enforcing
        mol3_reordered = dm.reorder_mol_from_template(
            mol3, mol3_variation, enforce_atomic_num=False, enforce_bond_type=False
        )
        self.assertIsInstance(mol3_reordered, dm.Mol)
        atoms3 = [atom.GetAtomicNum() for atom in mol3.GetAtoms()]
        atoms3_ordered = [atom.GetAtomicNum() for atom in mol3_variation.GetAtoms()]
        atoms3_reordered = [atom.GetAtomicNum() for atom in mol3_reordered.GetAtoms()]
        equal_atoms3 = [atoms3_reordered[ii] == atoms3_ordered[ii] for ii in range(len(atoms3))]
        equal_atoms3_wrong = [atoms3_reordered[ii] == atoms3[ii] for ii in range(len(atoms3))]
        self.assertEqual(len(atoms3) - sum(equal_atoms3), 1)
        self.assertGreater(len(atoms3) - sum(equal_atoms3_wrong), 1)

        # Check re-ordering of molecules that fail due to enforcing
        mol3_reordered = dm.reorder_mol_from_template(
            mol3, mol3_variation, enforce_atomic_num=True, enforce_bond_type=False
        )
        self.assertIsNone(mol3_reordered)
        mol3_reordered = dm.reorder_mol_from_template(
            mol3, mol3_variation, enforce_atomic_num=False, enforce_bond_type=True
        )
        self.assertIsNone(mol3_reordered)


if __name__ == "__main__":
    ut.main()



import datamol as dm


def test_freesolv():
    data = dm.data.freesolv()
    assert data.shape == (642, 4)
    assert list(data.columns) == ["iupac", "smiles", "expt", "calc"]


def test_cdk2():
    data = dm.data.cdk2()
    assert data.shape == (47, 12)
    assert list(data.columns) == [
        "smiles",
        "mol",
        "id",
        "Cluster",
        "MODEL.SOURCE",
        "MODEL.CCRATIO",
        "r_mmffld_Potential_Energy-OPLS_2005",
        "r_mmffld_RMS_Derivative-OPLS_2005",
        "b_mmffld_Minimization_Converged-OPLS_2005",
        "s_st_Chirality_1",
        "s_st_Chirality_2",
        "s_st_Chirality_3",
    ]


def test_solubility():
    data = dm.data.solubility()
    assert data.shape == (1282, 7)
    assert list(data.columns) == [
        "mol",
        "ID",
        "NAME",
        "SOL",
        "SOL_classification",
        "smiles",
        "split",
    ]



from typing import List
from typing import Optional
from typing import Union
from typing import Any

import functools

import numpy as np
from sklearn.metrics import pairwise_distances_chunked
from scipy.spatial import distance

import datamol as dm


def pdist(
    mols: List[Union[str, dm.Mol]],
    n_jobs: Optional[int] = 1,
    squareform: bool = True,
    **fp_args: Any,
) -> np.ndarray:
    """Compute the pairwise tanimoto distance between the fingerprints of all the
    molecules in the input set.

    Args:
        mols: list of molecules
        n_jobs: Number of jobs for parallelization. Let to 1 for no
            parallelization. Set to -1 to use all available cores.
        squareform: Whether to return in square form (matrix) or in a condensed
            form (1D vector).
        **fp_args: list of args to pass to `to_fp()`.

    Returns:
        dist_mat
    """

    fps = dm.parallelized(
        functools.partial(dm.to_fp, as_array=True, **fp_args),
        mols,
        n_jobs=n_jobs,
    )

    fps = np.array(fps)

    dist_mat = distance.pdist(fps, metric="jaccard")

    if squareform:
        dist_mat = distance.squareform(dist_mat, force="tomatrix")

    return dist_mat


def cdist(
    mols1: List[Union[str, dm.Mol]],
    mols2: List[Union[str, dm.Mol]],
    n_jobs: Optional[int] = 1,
    distances_chunk: bool = False,
    distances_chunk_memory: int = 1024,
    distances_n_jobs: int = -1,
    **fp_args: Any,
) -> np.ndarray:
    """Compute the tanimoto distance between the fingerprints of each pair of
    molecules of the two collections of inputs.

    Args:
        mols1: list of molecules.
        mols2: list of molecules.
        n_jobs: Number of jobs for fingerprint computation. Let to 1 for no
            parallelization. Set to -1 to use all available cores.
        distances_chunk: Whether to use chunked computation.
        distances_chunk_memory: Memory size in MB to use for chunked computation.
        distances_n_jobs: Number of jobs for parallelization.
        **fp_args: list of args to pass to `to_fp()`.

    Returns:
        distmat
    """

    fps1 = dm.parallelized(
        functools.partial(dm.to_fp, as_array=True, **fp_args),
        mols1,
        n_jobs=n_jobs,
    )

    fps2 = dm.parallelized(
        functools.partial(dm.to_fp, as_array=True, **fp_args),
        mols2,
        n_jobs=n_jobs,
    )

    fps1 = np.array(fps1).astype(bool)
    fps2 = np.array(fps2).astype(bool)

    if distances_chunk:
        distances = pairwise_distances_chunked(
            fps1,
            fps2,
            metric="jaccard",
            n_jobs=distances_n_jobs,
            working_memory=distances_chunk_memory,
        )
        distances = [i for i in distances]
        distances = np.vstack(distances)
    else:
        distances = distance.cdist(fps1, fps2, metric="jaccard")

    return distances



from typing import Dict
from typing import List
from typing import Any
from typing import Tuple
from typing import Optional

import collections
import itertools

from rdkit import Chem
from rdkit.Chem import rdFMCS
from rdkit.Chem import rdmolops
from rdkit.Chem import rdRGroupDecomposition

from rdkit.Chem.rdmolops import AdjustQueryParameters
from rdkit.Chem.rdmolops import AdjustQueryProperties

from rdkit.Chem.Scaffolds import MurckoScaffold

from ..types import Mol
from ..mol import keep_largest_fragment
from ..mol import fix_mol
from ..mol import to_mol
from ..mol import add_hs
from ..mol import remove_hs
from ..convert import to_smiles
from ..convert import to_smarts
from ..convert import from_smarts


def trim_side_chain(mol: Chem.rdchem.Mol, core, unwanted_side_chains):
    """Trim list of side chain from a molecule."""

    mol = add_hs(mol)

    match = mol.GetSubstructMatch(core)
    map2idx = {}
    map2nei = {}
    unwanted2map = {}
    for patt in unwanted_side_chains:
        unwanted2map[patt] = [a.GetAtomMapNum() for a in patt.GetAtoms() if a.GetAtomMapNum()]
    unwanted_mapping = list(itertools.chain.from_iterable(unwanted2map.values()))

    for atom in core.GetAtoms():
        num = atom.GetAtomMapNum()
        if num and num in unwanted_mapping:
            mol_atom_idx = match[atom.GetIdx()]
            map2idx[mol_atom_idx] = num
            nei_atoms = mol.GetAtomWithIdx(mol_atom_idx).GetNeighbors()
            map2nei[mol_atom_idx] = [n.GetIdx() for n in nei_atoms if n.GetIdx() in match]

    emol = Chem.EditableMol(mol)
    for atom_idx, atom_map in map2idx.items():
        dummy = Chem.rdchem.Atom("*")
        dummy.SetAtomMapNum(atom_map)
        nei_idx = map2nei.get(atom_idx, [None])[0]
        if nei_idx:
            bond = mol.GetBondBetweenAtoms(atom_idx, nei_idx)
            emol.RemoveBond(atom_idx, nei_idx)
            new_ind = emol.AddAtom(dummy)
            emol.AddBond(nei_idx, new_ind, bond.GetBondType())

    mol = emol.GetMol()
    mol = remove_hs(mol)
    query_param = AdjustQueryParameters()
    query_param.makeDummiesQueries = False
    query_param.adjustDegree = False
    query_param.aromatizeIfPossible = True
    for patt, _ in unwanted2map.items():
        cur_frag = fix_mol(patt)
        mol = Chem.DeleteSubstructs(mol, cur_frag, onlyFrags=True)

    return keep_largest_fragment(mol)


def fuzzy_scaffolding(
    mols: List[Chem.rdchem.Mol],
    enforce_subs: Optional[List[str]] = None,
    n_atom_cuttoff: int = 8,
    additional_templates: Optional[List[Mol]] = None,
    ignore_non_ring: bool = False,
    mcs_params: Optional[Dict[Any, Any]] = None,
) -> Tuple[set, Dict[str, dict], Dict[str, list]]:
    """Generate fuzzy scaffold with enforceable group that needs to appear
    in the core, forcing to keep the full side chain if required

    Args:
        mols: List of all molecules
        enforce_subs: List of substructure to enforce on the scaffold.
        n_atom_cuttoff: Minimum number of atom a core should have.
        additional_templates: Additional template to use to generate scaffolds.
        ignore_non_ring: Whether to ignore atom no in murcko ring system, even if they are in the framework.
        mcs_params: Arguments of MCS algorithm.

    Returns:
        - `set` - `scaffolds` - All found scaffolds in the molecules as valid smiles.
        - `Dict[Dict]` - `scaffold_infos` - Infos on the scaffold mapping, ignoring any side chain that had
                to be enforced. Key corresponds to generic scaffold smiles
                Values at ['smarts'] corresponds to smarts representation of the true scaffold (from MCS)
                Values at ['mols'] corresponds to list of molecules matching the scaffold
        - `Dict[List]` - `scaffold_to_group` - Map between each generic scaffold and the R-groups decomposition row.
    """

    # NOTE(hadim): consider parallelize this (if possible).
    # NOTE(hadim): consider refactoring this function in smaller reusable functions.

    if enforce_subs is None:
        enforce_subs = []

    if additional_templates is None:
        additional_templates = []

    if mcs_params is None:
        mcs_params = {}

    rg_params = rdRGroupDecomposition.RGroupDecompositionParameters()
    rg_params.removeAllHydrogenRGroups = True
    rg_params.removeHydrogensPostMatch = True
    rg_params.alignment = rdRGroupDecomposition.RGroupCoreAlignment.MCS
    rg_params.matchingStrategy = rdRGroupDecomposition.RGroupMatching.Exhaustive
    rg_params.rgroupLabelling = rdRGroupDecomposition.RGroupLabelling.AtomMap
    rg_params.labels = rdRGroupDecomposition.RGroupLabels.AtomIndexLabels

    core_query_param = AdjustQueryParameters()
    core_query_param.makeDummiesQueries = True
    core_query_param.adjustDegree = False
    core_query_param.makeBondsGeneric = True

    # group molecules by they generic Murcko scaffold, allowing
    # side chain that contains cycle (might be a bad idea)
    scf2infos = collections.defaultdict(dict)
    scf2groups = {}
    all_scaffolds = set([])

    for m in mols:
        generic_m = MurckoScaffold.MakeScaffoldGeneric(m)
        scf = MurckoScaffold.GetScaffoldForMol(m)
        try:
            scf = MurckoScaffold.MakeScaffoldGeneric(scf)
        except:
            pass

        if ignore_non_ring:
            rw_scf = Chem.RWMol(scf)
            atms = [a.GetIdx() for a in rw_scf.GetAtoms() if not a.IsInRing()]
            atms.sort(reverse=True)
            for a in atms:
                rw_scf.RemoveAtom(a)
            scfs = list(rdmolops.GetMolFrags(rw_scf, asMols=False))
        else:
            scfs = [to_smiles(scf)]

        # add templates mols if exists:
        for tmp in additional_templates:
            tmp = to_mol(tmp)
            tmp_scf = MurckoScaffold.MakeScaffoldGeneric(tmp)
            if generic_m.HasSubstructMatch(tmp_scf):
                scfs.append(to_smiles(tmp_scf))

        for scf in scfs:
            if scf2infos[scf].get("mols"):
                scf2infos[scf]["mols"].append(m)
            else:
                scf2infos[scf]["mols"] = [m]

    for scf in scf2infos:
        # cheat by adding murcko as last mol always
        popout = False
        mols = scf2infos[scf]["mols"]
        if len(mols) < 2:
            mols = mols + [MurckoScaffold.GetScaffoldForMol(mols[0])]
            popout = True

        # compute the MCS of the cluster
        mcs = rdFMCS.FindMCS(
            mols,
            atomCompare=rdFMCS.AtomCompare.CompareAny,
            bondCompare=rdFMCS.BondCompare.CompareAny,
            completeRingsOnly=True,
            **mcs_params,
        )

        mcsM = from_smarts(mcs.smartsString)
        mcsM.UpdatePropertyCache(False)
        Chem.SetHybridization(mcsM)

        if mcsM.GetNumAtoms() < n_atom_cuttoff:
            continue

        scf2infos[scf]["smarts"] = to_smarts(mcsM)
        if popout:
            mols = mols[:-1]

        core_groups = []
        # generate rgroups based on the mcs core
        success_mols = []
        try:
            rg = rdRGroupDecomposition.RGroupDecomposition(mcsM, rg_params)
            for i, analog in enumerate(mols):
                analog.RemoveAllConformers()
                res = rg.Add(analog)
                if not (res < 0):
                    success_mols.append(i)
            rg.Process()
            core_groups = rg.GetRGroupsAsRows()
        except Exception:
            pass

        mols = [mols[i] for i in success_mols]
        scf2groups[scf] = core_groups
        for mol, gp in zip(mols, core_groups):
            core = gp["Core"]
            acceptable_groups = [
                a.GetAtomMapNum()
                for a in core.GetAtoms()
                if (a.GetAtomMapNum() and not a.IsInRing())
            ]

            rgroups = [gp[f"R{k}"] for k in acceptable_groups if f"R{k}" in gp.keys()]
            if enforce_subs is not None:
                rgroups = [
                    rgp
                    for rgp in rgroups
                    if not any([len(rgp.GetSubstructMatch(frag)) > 0 for frag in enforce_subs])
                ]
            try:
                scaff = trim_side_chain(mol, AdjustQueryProperties(core, core_query_param), rgroups)
            except:
                continue
            all_scaffolds.add(to_smiles(scaff))

    return all_scaffolds, scf2infos, scf2groups



from typing import cast
from typing import Union
from typing import Sequence
from typing import Optional

import os
import io

from loguru import logger
import numpy as np

import fsspec

from rdkit.Chem import rdChemReactions

import datamol as dm


ATTACHING_RXN = rdChemReactions.ReactionFromSmarts("[*;h:1]>>[*:1][*]")


def rxn_from_smarts(rxn_smarts: str) -> dm.ChemicalReaction:
    """
    Create a reaction from smarts

    Args:
        rxn_smarts:  Reaction SMARTS string

    Returns:
        Initilized reaction.
    """
    rxn = rdChemReactions.ReactionFromSmarts(SMARTS=rxn_smarts)
    rxn.Initialize()
    return rxn


def rxn_to_smarts(rxn: dm.ChemicalReaction) -> str:
    """Create a SMARTS from a reaction.

    Args:
        rxn: dm.ChemicalReaction object.

    Returns:
        SMARTS as string.
    """
    return rdChemReactions.ReactionToSmarts(reaction=rxn)


def rxn_from_block(
    rxn_block: str,
    sanitize: bool = False,
) -> dm.ChemicalReaction:
    """Create a reaction from a block.

    Args:
        rxn_block: A reaction block.
        sanitize: Whether to sanitize the reaction.

    Returns:
        Initialized reaction.

    """
    rxn = rdChemReactions.ReactionFromRxnBlock(rxnblock=rxn_block, sanitize=sanitize)
    rxn.Initialize()
    return rxn


def rxn_from_block_file(
    rxn_block_path: Union[str, os.PathLike],
    sanitize: bool = False,
) -> dm.ChemicalReaction:
    """Create a reaction from a block file.

    Args:
        rxn_block_path: Filepath to a reaction block file.
        sanitize: Whether to sanitize the reaction.

    Returns:
        Initialized reaction.
    """
    with fsspec.open(rxn_block_path) as f:
        rxn_block = f.read()  # type: ignore
        rxn = rxn_from_block(rxn_block=rxn_block, sanitize=sanitize)
    return rxn


def rxn_to_block(
    rxn: dm.ChemicalReaction,
    separate_agents: bool = False,
    force_V3000: bool = False,
) -> str:
    """Create a block from a reaction object.

    Args:
        rxn: A reaction object.
        separate_agents: Whether to separate agents from the reactants block. Not supported
            if `force_V3000=True`.
        force_V3000: Write the block in a V3000 format.

    Returns:
        Reaction block as string
    """

    args = {}
    if dm.is_lower_than_current_rdkit_version("2022"):
        logger.warning("RDKit version prior to 2022.* does not support the `force_V3000` flag.")
    else:
        args["forceV3000"] = force_V3000

    return rdChemReactions.ReactionToRxnBlock(reaction=rxn, separateAgents=separate_agents, **args)


def rxn_to_block_file(
    rxn: dm.ChemicalReaction,
    output_block_path: Union[str, os.PathLike],
    separate_agents: bool = False,
    force_V3000: bool = False,
):
    """Create a block from a reaction object.

    Args:
        rxn: A reaction object.
        output_block_path: Filepath to a reaction block file.
        separate_agents: Whether to separate agents from the reactants block. Not supported
            if `force_V3000=True`.
        force_V3000: Write the block in a V3000 format.
    """
    block = rxn_to_block(
        rxn=rxn,
        separate_agents=separate_agents,
        force_V3000=force_V3000,
    )

    with fsspec.open(output_block_path, "w") as f:
        f = cast(io.TextIOWrapper, f)
        f.write(block)


def is_reaction_ok(rxn: dm.ChemicalReaction, enable_logs: bool = False) -> bool:
    """Check if the given reaction is synthetically valid.

    Args:
        rxn: dm.ChemicalReaction object
        enable_logs: Whether to enable logs.

    Returns:
        Boolean whether reaction is valid
    """
    nWarn, nError, nReactants, nProducts, labels = rdChemReactions.PreprocessReaction(rxn)

    if enable_logs:
        logger.info(f"Number of warnings:{nWarn}")
        logger.info(f"Number of preprocessing errors: {nError}")
        logger.info(f"Number of reactants in reaction: {nReactants}")
        logger.info(f"Number of products in reaction: {nProducts}")
        logger.info(f"Preprocess labels added:{labels}")

    return rdChemReactions.SanitizeRxn(rxn) in [
        rdChemReactions.SanitizeFlags.SANITIZE_NONE,
        None,
    ]


def select_reaction_output(
    product: Sequence[Sequence[dm.Mol]],
    product_index: Optional[Union[int, list]] = None,
    single_product_group: bool = True,
    rm_attach: bool = False,
    as_smiles: bool = False,
    sanitize: bool = True,
) -> Union[list, str, dm.Mol]:
    """
    Compute the products from a reaction. It only takes the first product of the

    Args:
        product: All the products from a reaction. A sequence of the list of products.
        product_index: Index of the product to select.
            Examples: A.B -> C.D. The indices of products are 0 and 1.
            Both C and D will be returned if index is None or product indices are to [0, 1].
        single_product_group: Whether return a single group of products from a reaction.
        rm_attach: Whether remove the attachment point from the products.
        as_smiles: Whether return the result in smiles.
        sanitize: Whether sanitize the products to return.

    Returns:
        Processed products from reaction.
    """
    if len(product) == 0:
        return list(product)
    product = np.array(product)
    if product_index is not None:
        product = product[:, product_index]
    if single_product_group:
        index = np.random.randint(product.shape[0], size=1)
        product = product[index]
    if sanitize:
        product = np.vectorize(dm.sanitize_mol)(product)
    if rm_attach:
        fn = lambda x: dm.remove_dummies(x) if x is not None else x
        product = np.vectorize(fn)(product)
    if as_smiles:
        fn = lambda x: dm.to_smiles(x, allow_to_fail=True) if x is not None else x
        product = np.vectorize(fn)(product)
    product = product.tolist()
    if single_product_group:
        return product[0]
    return product


def apply_reaction(
    rxn: dm.ChemicalReaction,
    reactants: tuple,
    product_index: Optional[Union[int, list]] = None,
    single_product_group: bool = False,
    as_smiles: bool = False,
    rm_attach: bool = False,
    disable_logs: bool = True,
    sanitize: bool = True,
) -> Union[list, str, dm.Mol]:
    """
    Apply a chemical reaction on a molecule

    Args:
       rxn: Reaction object.
       reactants: A tuple of reactants.
       product_index: The index of the product of interest.
       single_product_group: Whether return one product group from all possible product groups.
       as_smiles: Whether return products in SMILES.
       rm_attach: Whether remove the attachment point from products.
       disable_logs: Whether disable rdkit logs.
       sanitize: Whether sanitize the products.

    Returns:
       Reaction products.
    """
    with dm.without_rdkit_log(enable=disable_logs):
        if not rxn.IsInitialized():
            rxn.Initialize()  # pragma: no cover

        product = rxn.RunReactants(reactants)
        outputs = select_reaction_output(
            product=product,
            product_index=product_index,
            single_product_group=single_product_group,
            as_smiles=as_smiles,
            rm_attach=rm_attach,
            sanitize=sanitize,
        )

    return outputs


def can_react(rxn: dm.ChemicalReaction, mol: dm.Mol) -> bool:
    """Check if a molecule is a reactant to a chemical reaction.

    Args:
        rxn: Reaction to check.
        mol: Molecule to check if it is a reactant.

    Returns:
        True if `mol` is a reactant of rxn.
    """
    if not rxn.IsInitialized():
        rxn.Initialize()  # pragma: no cover
    if rxn.IsMoleculeReactant(mol):
        return find_reactant_position(rxn, mol) != -1
    return False


def find_reactant_position(rxn: dm.ChemicalReaction, mol: dm.Mol) -> int:
    """Find the position of a reactant in a reaction.

    Args:
        rxn: Reaction
        mol: Molecule

    Returns:
        Reactant position or -1 if `mol` is not a reactant.
    """

    if not rxn.IsInitialized():
        rxn.Initialize()  # pragma: no cover

    react_pos = -1
    for pos, rct in enumerate(rxn.GetReactants()):
        if mol.HasSubstructMatch(rct):
            react_pos = pos
    return react_pos


def inverse_reaction(rxn: dm.ChemicalReaction) -> dm.ChemicalReaction:
    """
    Get the reverse reaction of the input reaction

    Args:
        rxn: Reaction to inverse.

    Returns:
        Inversed reaction.
    """
    rxn2 = rdChemReactions.ChemicalReaction()
    for i in range(rxn.GetNumReactantTemplates()):
        rxn2.AddProductTemplate(rxn.GetReactantTemplate(i))
    for i in range(rxn.GetNumProductTemplates()):
        rxn2.AddReactantTemplate(rxn.GetProductTemplate(i))
    rxn2.Initialize()
    return rxn2



from typing import List
from typing import Tuple
from typing import Any
from typing import Optional

import io
import math
import random

from matplotlib.font_manager import FontManager

from rdkit import Chem
from rdkit.Chem import Draw
from PIL import Image, ImageDraw, ImageFont, ImageOps


def circle_grid(
    center_mol: Chem.rdchem.Mol,
    circle_mols: List[List[Chem.rdchem.Mol]],
    legend: Optional[str] = None,
    mol_size: Tuple[int, int] = (200, 200),
    circle_margin: int = 50,
    act_mapper: Optional[dict] = None,
):
    """Show molecules in concentric rings, with one molecule at the center

    Args:
        center_mol (Chem.Mol): Molecule at center
        circle_mols (list of list of <Chem.Mol>): List of molecule for each concentric circle around the center mol
        mol_size (tuple, optional): Tuple of width and height for each molecule
        circle_margin (int, optional): Margin between the circle layers
        act_mapper (dict): Map each molecule to a dictionary of activity
    """
    return MolsCircleGrid(center_mol, circle_mols, legend, mol_size, circle_margin, act_mapper)


class MolsCircleGrid:
    def __init__(
        self,
        center_mol: Chem.rdchem.Mol,
        circle_mols: List[List[Chem.rdchem.Mol]],
        legend: Optional[str] = None,
        mol_size: Tuple[int, int] = (200, 200),
        circle_margin: int = 50,
        act_mapper: Optional[dict] = None,
    ):
        """Show molecules in concentric rings, with one molecule at the center

        Args:
            center_mol: Molecule at center
            circle_mols: List of molecule for each concentric circle around the center mol
            mol_size: Tuple of width and height for each molecule
            circle_margin: Margin between the circle layers
            act_mapper: Map each molecule to a dictionary of activity
        """
        self.circle_mols = circle_mols
        self.circle_count = len(self.circle_mols)
        self.legend = legend or ""
        self.margin = circle_margin
        self.center_mol = center_mol
        self.mol_size = mol_size
        size = (max(mol_size) + self.margin) * (self.circle_count + 1)
        self.size = size
        self.image = Image.new(mode="RGBA", size=(size, size), color=(255, 255, 255, 0))
        self.midpoint = size // 2
        self.draw = None
        self.act_mapper = act_mapper or {}
        self._draw()

    def show(self, crop=False):
        if crop:
            crop_img = ImageOps.crop(self.image, border=1)
        else:
            crop_img = self.image
        return crop_img.show()

    def save(self, filename):
        self.image.save(filename)

    def _draw(self):
        """Create circles and slices in-memory"""
        draw = ImageDraw.Draw(self.image)
        self.draw = draw
        all_radius = self._draw_circles(draw)
        self._draw_center_mol()
        self._draw_ring_mols(all_radius)
        font = None
        w, h = draw.textsize(self.legend)
        try:
            fn = FontManager()
            fontpath = fn.findfont("Droid sans")
            font = ImageFont.truetype(fontpath, 12 * self.size // 800)
            w, h = font.getsize(self.legend)
        except:
            pass
        draw.text(
            ((self.size // 2 - w) - 2, self.size - 2 * h),
            self.legend,
            fill="black",
            font=font,
        )
        del draw
        self.draw = None

    def _repr_png_(self):
        bio = io.BytesIO()
        self.image.save(bio, format="PNG")
        return bio.getvalue()

    def _draw_circles(self, draw):
        if self.circle_count <= 0:
            return []
        radius_step = int(self.midpoint / (self.circle_count + 1))
        radius_list = []
        full_range = range(0, self.size // 2, radius_step)

        for i, radius in enumerate(full_range):
            radius += self.margin // 2
            bounding_box = [
                (self.midpoint - radius, self.midpoint - radius),
                (self.midpoint + radius, self.midpoint + radius),
            ]
            if radius > self.margin:
                transp = int(255 - (200 * (i - 1) / len(full_range)))
                draw.arc(bounding_box, 0, 360, fill=(190, 190, 190, transp))
            radius_list.append(radius + radius_step)
        return radius_list

    def _draw_mol_at(
        self,
        mol,
        center_x,
        center_y,
        mol_size=None,
        act_dict={},
        center=False,
        **kwargs: Any,
    ):
        img = mol
        if mol_size is None:
            mol_size = self.mol_size

        if isinstance(mol, Chem.Mol):
            img = Draw.MolToImage(mol, mol_size, kekulize=True, fitImage=True, **kwargs)

        width, height = img.size
        self.image.paste(img, (int(center_x - width / 2), int(center_y - height / 2)))
        txt = []
        for prop, propval in act_dict.items():
            if not isinstance(propval, str):
                propval = "{:.2f}".format(propval)
            txt.append(f"{prop}: {propval}")
        if txt and self.draw is not None:
            txt = "\n".join(txt)
            font = None
            w, h = self.draw.multiline_textsize(txt)
            try:
                fn = FontManager()
                fontpath = fn.findfont("Droid sans")
                font = ImageFont.truetype(fontpath, 18 + center * 8)
                w, h = self.draw.multiline_textsize(txt, font=font)
            except:
                pass

    def _draw_center_mol(self):
        self._draw_mol_at(
            self.center_mol,
            self.midpoint,
            self.midpoint,
            mol_size=[x + self.margin for x in self.mol_size],
            act_dict=self.act_mapper.get(self.center_mol, {}),
            center=True,
        )

    def _draw_ring_mols(self, radius_list):
        for i, mols in enumerate(self.circle_mols):
            radius = radius_list[i]
            ni = len(mols)
            rand_unit = random.random() * 2 * math.pi
            for k, mol in enumerate(mols):
                center_x = radius * math.cos(2 * k * math.pi / ni + rand_unit) + self.midpoint
                center_y = radius * math.sin(2 * k * math.pi / ni + rand_unit) + self.midpoint
                self._draw_mol_at(mol, center_x, center_y, act_dict=self.act_mapper.get(mol, {}))
